"""Implementation of IIIF image manipulations using netpbm programs.

This manipulator really is very very slow, it writes intermediate files
in order to keep the manipulations modular. Starting from a JPEG seems
especially slow. Strictly for play...
"""

import re
import os
import os.path
import glob
import magic
import subprocess

from .error import IIIFError
from .request import IIIFRequest
from .manipulator import IIIFManipulator


class IIIFManipulatorNetpbm(IIIFManipulator):
    """Class to manipulate an image with netpbm according to IIIF.

    All exceptions are raised as IIIFError objects which directly
    determine the HTTP response.
    """

    tmpdir = '/tmp'
    pnmdir = None

    def __init__(self, **kwargs):
        """Initialize IIIFManipulatorNetpbm object.

        Keyword arguments are passes to superclass initialize method.
        """
        super(IIIFManipulatorNetpbm, self).__init__(**kwargs)
        self.complianceLevel = "http://iiif.example.org/compliance/level/1"
        if (self.pnmdir is None):
            self.find_binaries()

    @classmethod
    def find_binaries(cls, tmpdir=None, shellsetup=None, pnmdir=None):
        """Set instance variables for directory and binary locations.

        FIXME - should accept params to set things other than defaults.
        """
        cls.tmpdir = ('/tmp' if (tmpdir is None) else tmpdir)
        # Shell setup command (e.g set library path)
        cls.shellsetup = ('' if (shellsetup is None) else shellsetup)
        if (pnmdir is None):
            cls.pnmdir = '/usr/bin'
            for dir in ('/usr/local/bin', '/sw/bin'):
                if (os.path.isfile(os.path.join(dir, 'pngtopnm'))):
                    cls.pnmdir = dir
        else:
            cls.pnmdir = pnmdir
        # Recklessly assume everything else under cls.pnmdir
        cls.pngtopnm = os.path.join(cls.pnmdir, 'pngtopnm')
        cls.jpegtopnm = os.path.join(cls.pnmdir, 'jpegtopnm')
        cls.pnmfile = os.path.join(cls.pnmdir, 'pnmfile')
        cls.pnmcut = os.path.join(cls.pnmdir, 'pnmcut')
        cls.pnmscale = os.path.join(cls.pnmdir, 'pnmscale')
        cls.pnmrotate = os.path.join(cls.pnmdir, 'pnmrotate')
        cls.pnmflip = os.path.join(cls.pnmdir, 'pnmflip')
        cls.pnmtopng = os.path.join(cls.pnmdir, 'pnmtopng')
        cls.ppmtopgm = os.path.join(cls.pnmdir, 'ppmtopgm')
        cls.pnmtotiff = os.path.join(cls.pnmdir, 'pnmtotiff')
        cls.pnmtojpeg = os.path.join(cls.pnmdir, 'pnmtojpeg')
        cls.pamditherbw = os.path.join(cls.pnmdir, 'pamditherbw')
        # Need djatoka to get jp2 output
        cls.djatoka_comp = '/Users/simeon/packages/adore-djatoka-1.1/bin/compress.sh'

    def do_first(self):
        """Create PNM file from input image file."""
        pid = os.getpid()
        self.basename = os.path.join(self.tmpdir, 'iiif_netpbm_' + str(pid))
        outfile = self.basename + '.pnm'
        # Convert source file to pnm
        filetype = self.file_type(self.srcfile)
        if (filetype == 'png'):
            if (self.shell_call(self.pngtopnm + ' ' + self.srcfile + ' > ' + outfile)):
                raise IIIFError(text="Oops... got error from pngtopnm.")
        elif (filetype == 'jpg'):
            if (self.shell_call(self.jpegtopnm + ' ' + self.srcfile + ' > ' + outfile)):
                raise IIIFError(text="Oops... got error from jpegtopnm.")
        else:
            raise IIIFError(code='501',
                            text='bad input file format (only know how to read png/jpeg)')
        self.tmpfile = outfile
        # Get size
        (self.width, self.height) = self.image_size(self.tmpfile)

    def do_region(self, x, y, w, h):
        """Apply region selection."""
        infile = self.tmpfile
        outfile = self.basename + '.reg'
        # simeon@ice ~>cat m.pnm | pnmcut 10 10 100 200 > m1.pnm
        if (x is None):
            # print "region: full"
            self.tmpfile = infile
        else:
            # print "region: (%d,%d,%d,%d)" % (x,y,w,h)
            if (self.shell_call('cat ' + infile + ' | ' + self.pnmcut + ' ' + str(x) + ' ' + str(y) + ' ' + str(w) + ' ' + str(h) + '  > ' + outfile)):
                raise IIIFError(text="Oops... got nonzero output from pnmcut.")
            self.width = w
            self.height = h
            self.tmpfile = outfile

    def do_size(self, w, h):
        """Apply size scaling."""
        # simeon@ice ~>cat m1.pnm | pnmscale -width 50 > m2.pnm
        infile = self.tmpfile
        outfile = self.basename + '.siz'
        if (w is None):
            # print "size: no scaling"
            self.tmpfile = infile
        else:
            # print "size: scaling to (%d,%d)" % (w,h)
            if (self.shell_call('cat ' + infile + ' | ' + self.pnmscale + ' -width ' + str(w) + ' -height ' + str(h) + '  > ' + outfile)):
                raise IIIFError(
                    text="Oops... got nonzero output from pnmscale.")
            self.width = w
            self.height = h
            self.tmpfile = outfile

    def do_rotation(self, mirror, rot):
        """Apply rotation and/or mirroring."""
        infile = self.tmpfile
        outfile = self.basename + '.rot'
        # NOTE: pnmrotate: angle must be between -90 and 90 and
        # rotations is CCW not CW per IIIF spec
        #
        # BUG in pnmrotate: +90 and -90 rotations the output image
        # size may be off. See for example a 1000x1000 image becoming
        # 1004x1000:
        #
        # simeon@RottenApple iiif>file testimages/67352ccc-d1b0-11e1-89ae-279075081939.png
        # testimages/67352ccc-d1b0-11e1-89ae-279075081939.png: PNG image data, 1000 x 1000, 8-bit/color RGB, non-interlaced
        # simeon@RottenApple iiif>cat testimages/67352ccc-d1b0-11e1-89ae-279075081939.png  | pngtopnm | pnmrotate -90 | pnmtopng > a.png; file a.png; rm a.png
        # a.png: PNG image data, 1004 x 1000, 8-bit/color RGB, non-interlaced
        # simeon@RottenApple iiif>cat testimages/67352ccc-d1b0-11e1-89ae-279075081939.png  | pngtopnm | pnmrotate 90 | pnmtopng > a.png; file a.png; rm a.png
        # a.png: PNG image data, 1004 x 1000, 8-bit/color RGB, non-interlaced
        #
        # WORKAROUND is to add a pnmscale for the 90degree case, some
        # simeon@RottenApple iiif>cat testimages/67352ccc-d1b0-11e1-89ae-279075081939.png  | pngtopnm | pnmrotate -90| pnmscale -width 1000 -height 1000 | pnmtopng > a.png; file a.png; rm a.png
        # a.png: PNG image data, 1000 x 1000, 8-bit/color RGB, non-interlaced
        #
        # FIXME - add mirroring
        #
        if (rot == 0.0):
            # print "rotation: no rotation"
            self.tmpfile = infile
        elif (rot <= 90.0 or rot >= 270.0):
            if (rot >= 270.0):
                rot -= 360.0
            # print "rotation: by %f degrees clockwise" % (rot)
            if (self.shell_call('cat ' + infile + ' | ' + self.pnmrotate + ' -background=#FFF ' + str(-rot) + '  > ' + outfile)):
                raise IIIFError(
                    text="Oops... got nonzero output from pnmrotate.")
            self.tmpfile = outfile
        else:
            # Between 90 and 270 = flip and then -90 to 90
            rot -= 180.0
            # print "rotation: by %f degrees clockwise" % (rot)
            if (self.shell_call('cat ' + infile + ' | ' + self.pnmflip + ' -rotate180 | ' + self.pnmrotate + ' ' + str(-rot) + '  > ' + outfile)):
                raise IIIFError(
                    text="Oops... got nonzero output from pnmrotate.")
            self.tmpfile = outfile
        # Fixup size for 90s
        if (abs(rot % 180.0 - 90.0) < 0.001):
            outfile2 = self.basename + '.rot2'
            if (self.shell_call('cat ' + self.tmpfile + ' | ' + self.pnmscale + ' -width ' + str(self.height) + ' -height ' + str(self.width) + ' > ' + outfile2)):
                raise IIIFError(
                    text="Oops... failed to fixup size after pnmrotate.")
            self.tmpfile = outfile2

    def do_quality(self, quality):
        """Apply value of quality parameter."""
        infile = self.tmpfile
        outfile = self.basename + '.col'
        # Quality (bit-depth):
        if (quality == 'grey' or quality == 'gray'):
            if (self.shell_call('cat ' + infile + ' | ' + self.ppmtopgm + ' > ' + outfile)):
                raise IIIFError(
                    text="Oops... got nonzero output from ppmtopgm.")
            self.tmpfile = outfile
        elif (quality == 'bitonal'):
            if (self.shell_call('cat ' + infile + ' | ' + self.ppmtopgm + ' | ' + self.pamditherbw + ' > ' + outfile)):
                raise IIIFError(
                    text="Oops... got nonzero output from ppmtopgm.")
            self.tmpfile = outfile
        elif ((quality == 'native' and self.api_version < '2.0') or
              (quality == 'default' and self.api_version >= '2.0') or
              quality == 'color'):
            self.tmpfile = infile
        else:
            raise IIIFError(code=400, parameter='quality',
                            text="Unknown quality parameter value requested.")

    def do_format(self, format):
        """Apply format selection."""
        infile = self.tmpfile
        outfile = self.basename + '.out'
        outfile_jp2 = self.basename + '.jp2'
        # Now convert finished pnm file to output format
        # simeon@ice ~>cat m3.pnm | pnmtojpeg  > m4.jpg
        # simeon@ice ~>cat m3.pnm | pnmtotiff > m4.jpg
        # pnmtotiff: computing colormap...
        # pnmtotiff: Too many colors - proceeding to write a 24-bit RGB file.
        # pnmtotiff: If you want an 8-bit palette file, try doing a 'ppmquant 256'.
        # simeon@ice ~>cat m3.pnm | pnmtopng  > m4.png
        fmt = ('png' if (format is None) else format)
        if (fmt == 'png'):
            # print "format: png"
            if (self.shell_call(self.pnmtopng + ' ' + infile + ' > ' + outfile)):
                raise IIIFError(
                    text="Oops... got nonzero output from pnmtopng.")
            mime_type = "image/png"
        elif (fmt == 'jpg'):
            # print "format: jpg"
            if (self.shell_call(self.pnmtojpeg + ' ' + infile + ' > ' + outfile)):
                raise IIIFError(
                    text="Oops... got nonzero output from pnmtojpeg.")
            mime_type = "image/jpeg"
        elif (fmt == 'tiff' or fmt == 'jp2'):
            # print "format: tiff/jp2"
            if (self.shell_call(self.pnmtotiff + ' ' + infile + ' > ' + outfile)):
                raise IIIFError(
                    text="Oops... got nonzero output from pnmtotiff.")
            mime_type = "image/tiff"
            if (fmt == 'jp2'):
                # use djatoka after tiff
                if (self.shell_call(DJATOKA_COMP + ' -i ' + outfile + ' -o ' + outfile_jp2)):
                    raise IIIFError(
                        text="Oops... got nonzero output from DJATOKA_COMP.")
                mime_type = "image/jp2"
                outfile = tmpfile_jp2
        else:
            raise IIIFError(code=415, parameter='format',
                            text="Unsupported output file format (%s), only png,jpg,tiff are supported." % (fmt))
        self.outfile = outfile
        self.output_format = fmt
        self.mime_type = mime_type

    def file_type(self, file):
        """Use python-magic to determine file type.

        Returns 'png' or 'jpg' on success, nothing on failure.
        """
        try:
            magic_text = magic.from_file(file).decode('utf-8')
        except (TypeError, IOError):
            return
        if (re.search('PNG image data', magic_text)):
            return('png')
        elif (re.search('JPEG image data', magic_text)):
            return('jpg')
        # failed
        return

    def image_size(self, pnmfile):
        """Get width and height of pnm file.

        simeon@homebox src>pnmfile /tmp/214-2.png
        /tmp/214-2.png:PPM raw, 100 by 100  maxval 255
        """
        pout = os.popen(self.shellsetup + self.pnmfile + ' ' + pnmfile, 'r')
        pnmfileout = pout.read(200)
        pout.close()
        m = re.search(', (\d+) by (\d+) ', pnmfileout)
        if (m is None):
            raise IIIFError(
                text="Bad output from pnmfile when trying to get size.")
        w = int(m.group(1))
        h = int(m.group(2))
        # print "pnmfile output = %s" % (pnmfileout)
        # print "image size = %d,%d" % (w,h)
        return(w, h)

    def shell_call(self, shellcmd):
        """Shell call with necessary setup first."""
        return(subprocess.call(self.shellsetup + shellcmd, shell=True))

    def cleanup(self):
        """Clean up any temporary files."""
        for file in glob.glob(self.basename + '*'):
            os.unlink(file)



"""Generator for IIIF Image API using the Python Image Library."""

import os.path
import sys

from PIL import Image

from .error import IIIFError
from .manipulator_pil import IIIFManipulatorPIL


class IIIFManipulatorGen(IIIFManipulatorPIL):
    """Class to generate an image with PIL according to IIIF Image API.

    Implemented as a small set of modifications to patch the generaton
    into the IIIFManipulatorPIL class which already handles rotation,
    quality and format parameters, and saving the image file.

    All exceptions are raised as IIIFError objects which directly
    determine the HTTP response.
    """

    def __init__(self, **kwargs):
        """Initialize IIIFManipulatorGen object.

        Keyword arguments are passed to superclass initialize method
        """
        super(IIIFManipulatorGen, self).__init__(**kwargs)
        self.gen = None

    def do_first(self):
        """Load generator, set size.

        We take the generator module name from self.srcfile so that
        this manipulator will work with different generators in a
        similar way to how the ordinary generators work with
        different images
        """
        # Load generator module and create instance if we haven't already
        if (not self.srcfile):
            raise IIIFError(text=("No generator specified"))
        if (not self.gen):
            try:
                (name, ext) = os.path.splitext(self.srcfile)
                (pack, mod) = os.path.split(name)
                module_name = 'iiif.generators.' + mod
                try:
                    module = sys.modules[module_name]
                except KeyError:
                    self.logger.debug(
                        "Loading generator module %s" % (module_name))
                    # Would be nice to use importlib but this is available only
                    # in python 2.7 and higher
                    pack = __import__(module_name)  # returns iiif package
                    module = getattr(pack.generators, mod)
                self.gen = module.PixelGen()
            except ImportError:
                raise IIIFError(
                    text=("Failed to load generator %s" % (str(self.srcfile))))
        (self.width, self.height) = self.gen.size

    def do_region(self, x, y, w, h):
        """Record region."""
        if (x is None):
            self.rx = 0
            self.ry = 0
            self.rw = self.width
            self.rh = self.height
        else:
            self.rx = x
            self.ry = y
            self.rw = w
            self.rh = h

    def do_size(self, w, h):
        """Record size."""
        if (w is None):
            self.sw = self.rw
            self.sh = self.rh
        else:
            self.sw = w
            self.sh = h
        # Now we have region and size, generate the image
        image = Image.new("RGB", (self.sw, self.sh), self.gen.background_color)
        for y in range(0, self.sh):
            for x in range(0, self.sw):
                ix = int((x * self.rw) // self.sw + self.rx)
                iy = int((y * self.rh) // self.sh + self.ry)
                color = self.gen.pixel(ix, iy)
                if (color is not None):
                    image.putpixel((x, y), color)
        self.image = image



#!/usr/bin/env python
"""iiif_static: Generate static images implementing the IIIF Image API level 0.

Copyright 2014--2016 Simeon Warner
"""

import logging
import optparse
import sys
import os.path

from iiif import __version__
from iiif.error import IIIFError
from iiif.static import IIIFStatic, IIIFStaticError


def main():
    """Parse arguments, instantiate IIIFStatic, run."""
    if (sys.version_info < (2, 6)):
        sys.exit("This program requires python version 2.6 or later")

    # Options and arguments
    p = optparse.OptionParser(description='IIIF Image API static file generator',
                              usage='usage: %prog [options] file [[file2..]] (-h for help)',
                              version='%prog ' + __version__)

    p.add_option('--dst', '-d', action='store', default='/tmp',
                 help="Destination directory for images [default '%default']")
    p.add_option('--tilesize', '-t', action='store', type='int', default=512,
                 help="Tilesize in pixels [default %default]")
    p.add_option('--api-version', '--api', '-a', action='store', default='2.1',
                 help="API version, may be 1.1, 2.0 or 2.1 [default %default]")
    p.add_option('--prefix', '-p', action='store', default=None,
                 help="URI prefix for where the images will be served from (default '%default'). "
                      "An empty prefix may be OK if the HTML page including the image shares the "
                      "the same root on the same server as the images, otherwise a full URL should "
                      "be specified. This is used to construct the @id in the info.json")
    p.add_option('--identifier', '-i', action='store', default=None,
                 help="Identifier for the image that will be used in place of the filename "
                      "(minus extension). Notes that this option cannot be used if more than "
                      "one image file is to be processed")
    p.add_option('--write-html', action='store', default=None,
                 help="Write HTML page to the specified directory using the 'identifier.html' "
                      "as the filename. HTML will launch OpenSeadragon for this image and to "
                      "display some of information about info.json and tile locations. HTML will "
                      "assume OpenSeadragon at relative path openseadragonVVV/openseadragon.min.js "
                      "and user-interface icons in openseadragonVVV/images, where VVV are the "
                      "three parts of the version number. The --include-osd flag is also specified "
                      "then OpenSeadragon will be copied to these locations")
    p.add_option('--include-osd', action='store_true',
                 help="Include OpenSeadragon files with --write-html flag")
    p.add_option('--osd-version', action='store', default='2.0.0',
                 help="Generate static images for older versions of OpenSeadragon. Use of versions "
                      "prior to 1.2.1 will force use of /w,h/ for size parameter instead of /w,/. "
                      "Likely useful only in combination with --api-version=1.1 "
                      "[default %default]")
    p.add_option('--osd-width', action='store', type='int', default='500',
                 help="Width of OpenSeadragon pane in pixels. Applies only with "
                      "--write-html [default %default]")
    p.add_option('--osd-height', action='store', type='int', default='500',
                 help="Height of OpenSeadragon pane in pixels. Applies only with "
                      "--write-html [default %default]")
    p.add_option('--generator', action='store_true', default=False,
                 help="Use named generator modules in iiif.generators package instead "
                      "of a starting image [default %default]")
    p.add_option('--max-image-pixels', action='store', type='int', default=0,
                 help="Set the maximum number of pixels in an image. A non-zero value "
                      "will set a hard limit on the image size. If left unset then the "
                      "default configuration of the Python Image Libary (PIL) will give "
                      "a DecompressionBombWarning if the image size exceeds a default "
                      "maximum, but otherwise continue as normal")
    p.add_option('--dryrun', '-n', action='store_true',
                 help="Do not write anything, say what would be done")
    p.add_option('--quiet', '-q', action='store_true',
                 help="Quite (no output unless there is a warning/error)")
    p.add_option('--verbose', '-v', action='store_true',
                 help="Verbose")

    (opt, sources) = p.parse_args()

    level = logging.DEBUG if (opt.verbose) else \
        logging.WARNING if (opt.quiet) else logging.INFO
    logging.basicConfig(format='%(name)s: %(message)s',
                        level=level)
    logger = logging.getLogger(os.path.basename(__file__))

    if (not opt.write_html and opt.include_osd):
        logger.warn(
            "--include-osd has no effect without --write-html, ignoring")
    if (len(sources) == 0):
        logger.warn("No sources specified, nothing to do, bye! (-h for help)")
    elif (len(sources) > 1 and opt.identifier):
        logger.error(
            "Cannot use --identifier/-i option with multiple sources, aborting.")
    else:
        try:
            sg = IIIFStatic(dst=opt.dst, tilesize=opt.tilesize,
                            api_version=opt.api_version, dryrun=opt.dryrun,
                            prefix=opt.prefix, osd_version=opt.osd_version,
                            generator=opt.generator,
                            max_image_pixels=opt.max_image_pixels)
            for source in sources:
                # File or directory (or neither)?
                if (os.path.isfile(source) or opt.generator):
                    logger.info("source file: %s" % (source))
                    sg.generate(source, identifier=opt.identifier)
                    if (opt.write_html):
                        sg.write_html(html_dir=opt.write_html, include_osd=opt.include_osd,
                                      osd_width=opt.osd_width, osd_height=opt.osd_height)
                elif (os.path.isdir(source)):
                    logger.warn(
                        "Ignoring source '%s': directory coversion not supported" % (source))
                else:
                    logger.warn(
                        "Ignoring source '%s': neither file nor path" % (source))
        except (IIIFStaticError, IIIFError) as e:
            # catch known errors and report nicely...
            logger.error("Error: " + str(e))

if __name__ == '__main__':
    main()



"""IIIF Authentication via Google.

FIXME - this code assumes Flask webapp framework, should be abstracted
"""

import json
import re
import os.path
try:
    # python3
    from urllib.request import urlopen
    from urllib.request import Request
    from urllib.parse import urlencode
except ImportError:
    # fall back to python2
    from urllib2 import urlopen
    from urllib2 import Request
    from urllib import urlencode
from flask import request, make_response, redirect

from .auth import IIIFAuth


class IIIFAuthGoogle(IIIFAuth):
    """IIIF Authentication Class using Google Auth."""

    def __init__(self, client_secret_file='client_secret.json', cookie_prefix=None):
        """Initialize IIIFAuthGoogle object.

        Sets a number of constants and attempts to load client client_secret
        data from client_secret_file.
        """
        super(IIIFAuthGoogle, self).__init__(cookie_prefix=cookie_prefix)
        #
        try:
            # Assign defaults so code/tests will have some data even if load
            # fails
            self.google_api_scope = 'https://www.googleapis.com/auth/userinfo.profile https://www.googleapis.com/auth/userinfo.email'
            self.google_oauth2_url = 'https://accounts.google.com/o/oauth2/'
            self.google_api_url = 'https://www.googleapis.com/oauth2/v1/'
            self.google_api_client_id = 'oops_missing_client_id'
            self.google_api_client_secret = 'oops_missing_client_secret'
            gcd = json.loads(open(client_secret_file).read())
            self.google_api_client_id = gcd['web']['client_id']
            self.google_api_client_secret = gcd['web']['client_secret']
        except Exception as e:
            self.logger.error("Failed to load Google auth from %s: %s" % (
                client_secret_file, str(e)))
        #
        # Auth data
        self.tokens = {}
        #
        self.account_cookie_name = self.cookie_prefix + 'account'

    def info_authn(self):
        """Check to see if user if authenticated for info.json.

        Must have Authorization header with value that is the appropriate
        token.
        """
        self.logger.info("info_authn: Authorization header = " +
                         request.headers.get('Authorization', '[none]'))
        return(request.headers.get('Authorization', '') != '')

    def image_authn(self):
        """Check to see if user if authenticated for image requests.

        Must have auth cookie with known token value.
        """
        self.logger.info("image_authn: auth cookie = " +
                         request.cookies.get(self.auth_cookie_name, default='[none]'))
        return request.cookies.get(self.auth_cookie_name, default='')

    def login_handler(self, config=None, prefix=None, **args):
        """OAuth starts here. This will redirect User to Google."""
        params = {
            'response_type': 'code',
            'client_id': self.google_api_client_id,
            'redirect_uri': self.scheme_host_port_prefix('http', config.host, config.port, prefix) + '/home',
            'scope': self.google_api_scope,
            'state': request.args.get('next', default=''),
        }
        url = self.google_oauth2_url + 'auth?' + urlencode(params)
        response = redirect(url)
        response.headers['Access-control-allow-origin'] = '*'
        return response

    def logout_handler(self, **args):
        """Handler for logout button.

        Delete cookies and return HTML that immediately closes window
        """
        response = make_response(
            "<html><script>window.close();</script></html>", 200, {'Content-Type': "text/html"})
        response.set_cookie(self.account_cookie_name, expires=0)
        response.set_cookie(self.auth_cookie_name, expires=0)
        response.headers['Access-Control-Allow-Origin'] = '*'
        return response

    def access_token_handler(self, **args):
        """Get access token based on cookie sent with this request.

        The client requests a token to send in re-request for info.json.
        Support JSONP request to get the token to send to info.json in
        Auth'z header.
        """
        callback_function = request.args.get('callback', default='')
        authcode = request.args.get('code', default='')
        account = request.cookies.get(self.account_cookie_name, default='')
        if (account):
            data = {"access_token": account,
                    "token_type": "Bearer", "expires_in": 3600}
        else:
            data = {"error": "client_unauthorized",
                    "error_description": "No login details received"}
        data_str = json.dumps(data)

        ct = "application/json"
        if (callback_function):
            data_str = "%s(%s);" % (callback_function, data_str)
            ct = "application/javascript"
        # Build response
        response = make_response(data_str, 200, {'Content-Type': ct})
        if (account):
            # Set the cookie for the image content -- FIXME - need something
            # real
            response.set_cookie(self.auth_cookie_name, account)
        response.headers['Access-control-allow-origin'] = '*'
        return response

    def home_handler(self, config=None, prefix=None, **args):
        """Handler for /home redirect path after Goole auth.

        OAuth ends up back here from Google. Set the account cookie
        and close window to trigger next step.
        """
        gresponse = self.google_get_token(config, prefix)
        gdata = self.google_get_data(config, gresponse)

        email = gdata.get('email', 'NO_EMAIL')
        name = gdata.get('name', 'NO_NAME')
        response = make_response(
            "<html><script>window.close();</script></html>", 200, {'Content-Type': "text/html"})
        response.set_cookie(self.account_cookie_name,
                            'Token for ' + name + ' ' + email)
        return response

    ######################################################################
    # Code to get data from Google API
    #

    def google_get_token(self, config, prefix):
        """Make request to Google API to get token."""
        params = {
            'code': request.args.get('code', default=''),
            'client_id': self.google_api_client_id,
            'client_secret': self.google_api_client_secret,
            'redirect_uri': self.scheme_host_port_prefix('http', config.host, config.port, prefix) + '/home',
            'grant_type': 'authorization_code',
        }
        payload = urlencode(params).encode('utf-8')
        url = self.google_oauth2_url + 'token'
        req = Request(url, payload)
        return json.loads(urlopen(req).read())

    def google_get_data(self, config, response):
        """Make request to Google API to get profile data for the user."""
        params = {
            'access_token': response['access_token'],
        }
        payload = urlencode(params)
        url = self.google_api_url + 'userinfo?' + payload
        req = Request(url)
        return json.loads(urlopen(req).read())



"""IIIF Image API error class.

In IIIF Image API version 1.1 and above the specification do not
mandate a particular format but suggest a human readable message.
Suitable output is provided in plain text by error.as_txt().

In IIIF Image API version 1.0 the specification mandates an XML
response following the format:

  <?xml version="1.0" encoding="UTF-8"?>
  <error xmlns="http://library.stanford.edu/iiif/image-api/ns/">
    <parameter>size</parameter>
    <text>Invalid size specified</text>
  </error>

This format is provided by error.as_xml().
"""

import sys
from xml.etree.ElementTree import ElementTree, Element
try:  # python2
    import BytesIO as io
    # Must check for python2 first as io exists but is wrong one
except ImportError:  # python3
    import io

# Namespace used in v1.0 XML error response
I3F_NS = "http://library.stanford.edu/iiif/image-api/ns/"


class IIIFError(Exception):
    """Class to represent IIIF error conditions."""

    def __init__(self, code=500, parameter='unknown', text='', headers=None):
        """Initialize IIIFError object.

        Keyword arguments:
        code -- HTTP error/status code, should always specify
        parameter -- parameter causing error to help with debugging
        test -- human explanation to help with debugging
        headers -- additional HTTP headers
        """
        self.code = code
        self.parameter = parameter
        self.text = text
        self.headers = headers if (headers is not None) else ()
        self.content_type = 'text/plain'
        self.pretty_xml = True

    def image_server_response(self, api_version=None):
        """Response, code and headers for image server error response.

        api_version selects the format (XML of 1.0). The return value is
        a tuple of
          response - body of HTTP response
          status - the HTTP status code
          headers - a dict of HTTP headers which will include the Content-Type

        As a side effect the routine sets self.content_type
        to the correct media type for the response.
        """
        headers = dict(self.headers)
        if (api_version < '1.1'):
            headers['Content-Type'] = 'text/xml'
            response = self.as_xml()
        else:
            headers['Content-Type'] = 'text/plain'
            response = self.as_txt()
        return(response, self.code, headers)

    def as_xml(self):
        """XML representation of the error to be used in HTTP response.

        This XML format follows the IIIF Image API v1.0 specification,
        see <http://iiif.io/api/image/1.0/#error>
        """
        # Build tree
        spacing = ("\n" if (self.pretty_xml) else "")
        root = Element('error', {'xmlns': I3F_NS})
        root.text = spacing
        e_parameter = Element('parameter', {})
        e_parameter.text = self.parameter
        e_parameter.tail = spacing
        root.append(e_parameter)
        if (self.text):
            e_text = Element('text', {})
            e_text.text = self.text
            e_text.tail = spacing
            root.append(e_text)
        # Write out as XML document to return
        tree = ElementTree(root)
        xml_buf = io.BytesIO()
        if (sys.version_info < (2, 7)):
            tree.write(xml_buf, encoding='UTF-8')
        else:
            tree.write(xml_buf, encoding='UTF-8',
                       xml_declaration=True, method='xml')
        return(xml_buf.getvalue().decode('utf-8'))

    def as_txt(self):
        """Text rendering of error response.

        Designed for use with Image API version 1.1 and above where the
        error response is suggested to be text or html but not otherwise
        specified. Intended to provide useful information for debugging.
        """
        s = "IIIF Image Server Error\n\n"
        s += self.text if (self.text) else 'UNKNOWN_ERROR'
        s += "\n\n"
        if (self.parameter):
            s += "parameter=%s\n" % self.parameter
        if (self.code):
            s += "code=%d\n\n" % self.code
        for header in sorted(self.headers):
            s += "header %s=%s\n" % (header, self.headers[header])
        return s

    def __str__(self):
        """Short human readable version of IIIF error.

        This rendering does not include the HTTP status code or header
        values in output. The intention is that image server implementations
        will use image_server_response(api_version) instead.
        """
        s = self.text if (self.text) else 'UNKNOWN_ERROR'
        if (self.parameter and self.parameter != 'unknown'):
            s += ", parameter=%s" % self.parameter
        return s


class IIIFZeroSizeError(IIIFError):
    """Sub-class of IIIFError to indicate request for a zero-size image."""

    pass



"""IIIF Image Information Response.

Model for IIIF Image API 'Image Information Response'.
Default version is 2.1 but also supports 2.0, 1.1 and 1.0.

Philisophy is to migrate this code forward with new versions
of the specification but to keep support for all published
versions.
"""

import sys
import json
import re


def _parse_int_array(info, json_data):
    # Force simple array of interger values
    return [int(x) for x in json_data]


def _parse_noop(info, json_data):
    # Format is already what we want
    return json_data


def _parse_tile(info, json_data):
    # Parse data for a single tile specification
    tile = {}
    tile['width'] = int(json_data['width'])
    if ('height' in json_data):
        tile['height'] = int(json_data['height'])
    tile['scaleFactors'] = _parse_int_array(info, json_data['scaleFactors'])
    return tile


def _parse_tiles(info, json_data):
    # Parse tiles array of tile specifications
    #
    # Expect common case in 2.0 to map to 1.1 idea of tile_width,
    # tile_height and scale_factors. This is the case when len()==1.
    tiles = []
    if (len(json_data) == 0):
        raise IIIFInfoError("Empty tiles array property not allowed.")
    for tile_spec in json_data:
        tiles.append(_parse_tile(info, tile_spec))
    return tiles


def _parse_service(info, json_data):
    return json_data


def _parse_profile(info, json_data):
    # 2.1 spec: "A list of profiles, indicated by either a URI or an
    # object describing the features supported. The first entry
    # in the list must be a compliance level URI."
    #
    # 2.0 spec: "An array of profiles, indicated by either a URI or
    # an object describing the features supported. The first entry in
    # the array must be a compliance level URI, as defined below."
    #
    # 1.1 & 1.0 spec: "URI indicating the compliance level supported.
    # Values as described in Section 8. Compliance Levels"
    #
    # FIXME - add some validation!
    return json_data


# Configuration information for API versions

CONF = {
    '1.0': {
        'params':
            ['identifier', 'protocol', 'width', 'height', 'scale_factors',
             'tile_width', 'tile_height', 'formats', 'qualities', 'profile'],
        'array_params': set(
            ['scale_factors', 'formats', 'qualities']),
        'complex_params': {
            'scale_factors': _parse_int_array,
            'formats': _parse_noop,  # array of str
            'qualities': _parse_noop},  # array of str
        'compliance_prefix':
            "http://library.stanford.edu/iiif/image-api/compliance.html#level",
        'compliance_suffix': "",
        'protocol': None,
        'required_params': ['identifier', 'width', 'height', 'profile'],
    },
    '1.1': {
        'params':
            ['identifier', 'protocol', 'width', 'height', 'scale_factors',
             'tile_width', 'tile_height', 'formats', 'qualities', 'profile'],
        'array_params': set(
            ['scale_factors', 'formats', 'qualities']),
        'complex_params': {
            'scale_factors': _parse_int_array,
            'formats': _parse_noop,  # array of str
            'qualities': _parse_noop},  # array of str
        'context':
            "http://library.stanford.edu/iiif/image-api/1.1/context.json",
        'compliance_prefix':
            "http://library.stanford.edu/iiif/image-api/1.1/compliance.html#level",
        'compliance_suffix': "",
        'protocol': None,
        'required_params': ['identifier', 'width', 'height', 'profile'],
    },
    '2.0': {
        'params':
            ['identifier', 'protocol', 'width', 'height', 'profile',
             'sizes', 'tiles', 'service'],
        # scale_factors isn't in API but used internally
        'array_params': set(
            ['sizes', 'tiles', 'service', 'scale_factors', 'formats',
             'qualities', 'supports']),
        'complex_params': {
            'sizes': _parse_noop,
            'tiles': _parse_tiles,
            'profile': _parse_profile,
            'service': _parse_service},
        'context': "http://iiif.io/api/image/2/context.json",
        'compliance_prefix': "http://iiif.io/api/image/2/level",
        'compliance_suffix': ".json",
        'protocol': "http://iiif.io/api/image",
        'required_params':
            ['identifier', 'protocol', 'width', 'height', 'profile'],
    },
    '2.1': {
        'params':
            ['identifier', 'protocol', 'width', 'height',
             'profile', 'sizes', 'tiles', 'service',
             'attribution', 'logo', 'license'],
        # scale_factors isn't in API but used internally
        'array_params': set(
            ['sizes', 'tiles', 'service', 'scale_factors', 'formats',
             'maxArea', 'maxHeight', 'maxWidth', 'qualities', 'supports']),
        'complex_params': {
            'sizes': _parse_noop,
            'tiles': _parse_tiles,
            'profile': _parse_profile,
            'service': _parse_service},
        'context': "http://iiif.io/api/image/2/context.json",
        'compliance_prefix': "http://iiif.io/api/image/2/level",
        'compliance_suffix': ".json",
        'protocol': "http://iiif.io/api/image",
        'required_params':
            ['identifier', 'protocol', 'width', 'height', 'profile'],
    }
}


class IIIFInfoError(Exception):
    """IIIFInfoErrors from IIIFInfo."""

    pass


class IIIFInfo(object):
    """IIIF Image Information Class."""

    def __init__(self, api_version='2.1', profile=None, level=1, conf=None,
                 server_and_prefix='',
                 identifier=None, width=None, height=None, tiles=None,
                 sizes=None, service=None, id=None,
                 # legacy params from 1.1
                 scale_factors=None, tile_width=None, tile_height=None,
                 # 1.1 and 2.0
                 formats=None, qualities=None,
                 # 2.0 only
                 supports=None,
                 # 2.1 only
                 attribution=None, logo=None, license=None
                 ):
        """Initialize an IIIFInfo object.

        Parameters include:
        api_version -- defaults to 2.0 but may be set to 1.1
        profile -- usually not set but overrides handling via level
            (2.0 complex profile not supported except by explicitly
            passing in an array that matches JSON required)
        level -- default 1, generates compliance level for simple profile
        """
        # API version (used in level settings)
        if (api_version not in CONF):
            raise IIIFInfoError(
                "Unknown IIIF Image API version '%s', versions supported are ('%s')" %
                (api_version, sorted(
                    CONF.keys())))
        self.api_version = api_version
        self.set_version_info()
        if (profile is not None):
            # Explicit profile setting overrides any level set
            # or default level
            self.profile = profile
        else:
            self.level = level
        # explicit settings
        self.server_and_prefix = server_and_prefix
        self.identifier = identifier
        self.width = width
        self.height = height
        self.tiles = tiles
        self.sizes = sizes
        self.service = service
        # legacy params for 1.0 and 1.1
        if (scale_factors is not None):
            self.scale_factors = scale_factors
        if (tile_width is not None):
            self.tile_width = tile_width
        if (tile_height is not None):
            self.tile_height = tile_height
        # 1.1+
        self.formats = formats
        self.qualities = qualities
        # 2.0+ only
        self.supports = supports
        # 2.1+ only
        self.attribution = attribution
        self.logo = logo
        self.license = license
        # defaults from conf dict if provided
        if (conf):
            for option in conf:
                setattr(self, option, conf[option])
        if (id is not None):
            self.id = id

    @property
    def id(self):
        """id property based on server_and_prefix and identifier."""
        id = ''
        if (self.server_and_prefix is not None and
                self.server_and_prefix != ''):
            id += self.server_and_prefix + '/'
        if (self.identifier is not None):
            id += self.identifier
        return id

    @id.setter
    def id(self, value):
        """Split into server_and_prefix and identifier."""
        i = value.rfind('/')
        if (i > 0):
            self.server_and_prefix = value[:i]
            self.identifier = value[(i + 1):]
        elif (i == 0):
            self.server_and_prefix = ''
            self.identifier = value[(i + 1):]
        else:
            self.server_and_prefix = ''
            self.identifier = value

    def set_version_info(self, api_version=None):
        """Set up normal values for given api_version.

        Will use current value of self.api_version if a version number
        is not specified in the call. Will raise an IIIFInfoError
        """
        if (api_version is None):
            api_version = self.api_version
        if (api_version not in CONF):
            raise IIIFInfoError("Unknown API version %s" % (api_version))
        self.params = CONF[api_version]['params']
        self.array_params = CONF[api_version]['array_params']
        self.complex_params = CONF[api_version]['complex_params']
        for a in ('context', 'compliance_prefix', 'compliance_suffix',
                  'protocol', 'required_params'):
            if (a in CONF[api_version]):
                self._setattr(a, CONF[api_version][a])

    @property
    def compliance(self):
        """Compliance profile URI.

        In IIIF Image API v2.x the profile information is an array
        of values and objects, the first of which must be the
        compliance level URI. In v1.x the profile information is
        just this URI.
        """
        if (self.api_version < '2.0'):
            return self.profile
        else:
            return self.profile[0]

    @compliance.setter
    def compliance(self, value):
        """Set the compliance profile URI."""
        if (self.api_version < '2.0'):
            self.profile = value
        else:
            try:
                self.profile[0] = value
            except AttributeError:
                # handle case where profile not initialized as array
                self.profile = [value]

    @property
    def level(self):
        """Extract level number from compliance profile URI.

        Returns integer level number or raises IIIFInfoError
        """
        m = re.match(
            self.compliance_prefix +
            r'(\d)' +
            self.compliance_suffix +
            r'$',
            self.compliance)
        if (m):
            return int(m.group(1))
        raise IIIFInfoError(
            "Bad compliance profile URI, failed to extract level number")

    @level.setter
    def level(self, value):
        """Build profile URI from level.

        Level should be an integer 0,1,2
        """
        self.compliance = self.compliance_prefix + \
            ("%d" % value) + self.compliance_suffix

    def _single_tile_getter(self, param):
        # Extract param from a single tileset defintion
        if (self.tiles is None or len(self.tiles) == 0):
            return None
        elif (len(self.tiles) == 1):
            return self.tiles[0].get(param, None)
        else:
            raise IIIFInfoError(
                "No single %s in the case of multiple tile definitions." % (param))

    def _single_tile_setter(self, param, value):
        # Set param for a single tileset defintion
        if (self.tiles is None or len(self.tiles) == 0):
            self.tiles = [{}]
        elif (len(self.tiles) > 1):
            raise IIIFInfoError(
                "No single %s in the case of multiple tile definitions." % (param))
        self.tiles[0][param] = value

    @property
    def scale_factors(self):
        """Access to scale_factors in 1.x.

        Also provides the scale factors in 2.0 and greater
        provided there is exactly one tiles definition.
        """
        return self._single_tile_getter('scaleFactors')

    @scale_factors.setter
    def scale_factors(self, value):
        """Set scale_factors at version 1.1 and below.

        If used at 2.0+ this will create/edit the first entry
        in self.tiles. Will throw and error if it is used with more
        than on entry for self.tiles.
        """
        self._single_tile_setter('scaleFactors', value)

    @property
    def tile_width(self):
        """Access to tile_width in 1.x.

        Also provides the tile_width in 2.0 and greater
        provided there is exactly one tiles definition.
        """
        return self._single_tile_getter('width')

    @tile_width.setter
    def tile_width(self, value):
        """Set tile_width at version 1.1 and below."""
        self._single_tile_setter('width', value)

    @property
    def tile_height(self):
        """Access to tile_height in 1.x.

        Also provides the tile_height in 2.0 and greater
        provided there is exactly one tiles definition. If
        width is set but not height then return that instead.
        """
        h = self._single_tile_getter('height')
        if (h is None):
            return self._single_tile_getter('width')
        return h

    @tile_height.setter
    def tile_height(self, value):
        """Set tile_height at version 1.1 and below."""
        self._single_tile_setter('height', value)

    def add_service(self, service):
        """Add a service description.

        Handles transition from self.service=None, self.service=dict for a
        single service, and then self.service=[dict,dict,...] for multiple
        """
        if (self.service is None):
            self.service = service
        elif (isinstance(self.service, dict)):
            self.service = [self.service, service]
        else:
            self.service.append(service)

    def _setattr(self, param, value):
        # Setter handling array/scalar properties, and legacy.
        #
        # Also deals with legacy parameters that map to other
        # variables (but ugly!).
        if (param in self.array_params and
                isinstance(value, str)):
            # If we have an array then set directly, make list. Perhaps not
            # pythonic to do a type check for array here but want to avoid
            # accidentally iterating on chars in string etc..
            value = [value]
        setattr(self, param, value)

    def validate(self):
        """Validate this object as Image API data.

        Raise IIIFInfoError with helpful message if not valid.
        """
        errors = []
        for param in self.required_params:
            if (not hasattr(self, param) or getattr(self, param) is None):
                errors.append("missing %s parameter" % (param))
        if (len(errors) > 0):
            raise IIIFInfoError("Bad data for info.json: " + ", ".join(errors))
        return True

    def as_json(self, validate=True):
        """Return JSON serialization.

        Will raise IIIFInfoError if insufficient parameters are present to
        have a valid info.json response (unless validate is False).
        """
        if (validate):
            self.validate()
        json_dict = {}
        if (self.api_version > '1.0'):
            json_dict['@context'] = self.context
        params_to_write = set(self.params)
        params_to_write.discard('identifier')
        if (self.identifier):
            if (self.api_version == '1.0'):
                json_dict['identifier'] = self.identifier  # local id
            else:
                json_dict['@id'] = self.id  # URI
        params_to_write.discard('profile')
        if (self.compliance):
            if (self.api_version < '2.0'):
                json_dict['profile'] = self.compliance
            else:
                # FIXME - need to support extra profile features
                json_dict['profile'] = [self.compliance]
                d = {}
                if (self.formats is not None):
                    d['formats'] = self.formats
                if (self.qualities is not None):
                    d['qualities'] = self.qualities
                if (self.supports is not None):
                    d['supports'] = self.supports
                if (len(d) > 0):
                    json_dict['profile'].append(d)
                params_to_write.discard('formats')
                params_to_write.discard('qualities')
                params_to_write.discard('supports')
        for param in params_to_write:
            if (hasattr(self, param) and
                    getattr(self, param) is not None):
                json_dict[param] = getattr(self, param)
        return(json.dumps(json_dict, sort_keys=True, indent=2))

    def read(self, fh, api_version=None):
        """Read info.json from file like object.

        Parameters:
        fh -- file like object supporting fh.read()
        api_version -- IIIF Image API version expected

        If api_version is set then the parsing will assume this API version,
        else the version will be determined from the incoming data. NOTE that
        the value of self.api_version is NOT used in this routine.
        If an api_version is specified and there is a @context specified then
        an IIIFInfoError will be raised unless these match. If no known
        @context is present and no api_version set then an IIIFInfoError
        will be raised.
        """
        j = json.load(fh)
        #
        # @context and API version
        self.context = None
        if (api_version == '1.0'):
            # v1.0 did not have a @context so we simply take the version
            # passed in
            self.api_version = api_version
        elif ('@context' in j):
            # determine API version from context
            self.context = j['@context']
            api_version_read = None
            for v in CONF:
                if (v > '1.0' and self.context == CONF[v]['context']):
                    api_version_read = v
                    break
            if (api_version_read is None):
                raise IIIFInfoError(
                    "Unknown @context, cannot determine API version (%s)" %
                    (self.context))
            else:
                if (api_version is not None and
                        api_version != api_version_read):
                    raise IIIFInfoError(
                        "Expected API version '%s' but got @context for API version '%s'" %
                        (api_version, api_version_read))
                else:
                    self.api_version = api_version_read
        else:  # no @context and not 1.0
            if (api_version is None):
                raise IIIFInfoError("No @context (and no default given)")
            self.api_version = api_version
        self.set_version_info()
        #
        # @id or identifier
        if (self.api_version == '1.0'):
            if ('identifier' in j):
                self.id = j['identifier']
            else:
                raise IIIFInfoError("Missing identifier in info.json")
        else:
            if ('@id' in j):
                self.id = j['@id']
            else:
                raise IIIFInfoError("Missing @id in info.json")
        #
        # other params
        for param in self.params:
            if (param == 'identifier'):
                continue  # dealt with above
            if (param in j):
                if (param in self.complex_params):
                    # use function ref in complex_params to parse, optional
                    # dst to map to a different property name
                    self._setattr(param, self.complex_params[
                                  param](self, j[param]))
                else:
                    self._setattr(param, j[param])
        return True



"""IIIF Image API module."""
from iiif._version import __version__

from iiif.error import IIIFError
from iiif.info import IIIFInfo
from iiif.manipulator import IIIFManipulator
from iiif.request import IIIFRequest
from iiif.static import IIIFStatic



"""Implementation of IIIF Image API manipulations using the Python Image Library.

Uses the Python Image Library (PIL) for in-python manipulation:
http://www.pythonware.com/products/pil/index.htm
"""

import re
import os
import os.path
import subprocess
import tempfile

from PIL import Image

from .error import IIIFError
from .request import IIIFRequest
from .manipulator import IIIFManipulator


class IIIFManipulatorPIL(IIIFManipulator):
    """Class to manipulate an image with PIL according to IIIF.

    All exceptions are raised as IIIFError objects which directly
    determine the HTTP response.
    """

    tmpdir = '/tmp'
    filecmd = None
    pnmdir = None

    def __init__(self, **kwargs):
        """Initialize IIIFManipulatorPIL object.

        Keyword arguments are passes to superclass initialize method.
        """
        super(IIIFManipulatorPIL, self).__init__(**kwargs)
        # Does not support jp2 output
        self.compliance_level = 2
        self.outtmp = None

    def set_max_image_pixels(self, pixels):
        """Set PIL limit on pixel size of images to load if non-zero.

        WARNING: This is a global setting in PIL, it is
        not local to this manipulator instance!

        Setting a value here will not only set the given limit but
        also convert the PIL "DecompressionBombWarning" into an
        error. Thus setting a moderate limit sets a hard limit on
        image size loaded, setting a very large limit will have the
        effect of disabling the warning.
        """
        if (pixels):
            Image.MAX_IMAGE_PIXELS = pixels
            Image.warnings.simplefilter(
                'error', Image.DecompressionBombWarning)

    def do_first(self):
        """Create PIL object from input image file.

        Image location must be in self.srcfile. Will result in
        self.width and self.height being set to the image dimensions.

        Will raise an IIIFError on failure to load the image
        """
        self.logger.debug("do_first: src=%s" % (self.srcfile))
        try:
            self.image = Image.open(self.srcfile)
        except Image.DecompressionBombWarning as e:
            # This exception will be raised only if PIL has been
            # configured to raise an error in the case of images
            # that exceeed Image.MAX_IMAGE_PIXELS, with
            # Image.warnings.simplefilter('error', Image.DecompressionBombWarning)
            raise IIIFError(text=("Image size limit exceeded (PIL: %s)" % (str(e))))
        except Exception as e:
            raise IIIFError(text=("Failed to read image (PIL: %s)" % (str(e))))
        (self.width, self.height) = self.image.size

    def do_region(self, x, y, w, h):
        """Apply region selection."""
        if (x is None):
            self.logger.debug("region: full (nop)")
        else:
            self.logger.debug("region: (%d,%d,%d,%d)" % (x, y, w, h))
            self.image = self.image.crop((x, y, x + w, y + h))
            self.width = w
            self.height = h

    def do_size(self, w, h):
        """Apply size scaling."""
        if (w is None):
            self.logger.debug("size: no scaling (nop)")
        else:
            self.logger.debug("size: scaling to (%d,%d)" % (w, h))
            self.image = self.image.resize((w, h))
            self.width = w
            self.height = h

    def do_rotation(self, mirror, rot):
        """Apply rotation and/or mirroring."""
        if (not mirror and rot == 0.0):
            self.logger.debug("rotation: no rotation (nop)")
        else:
            # FIXME - with PIL one can use the transpose() method to do 90deg
            # FIXME - rotations as well as mirroring. This would be more efficient
            # FIXME - for these cases than mirror _then_ rotate.
            if (mirror):
                self.logger.debug("rotation: mirror (about vertical axis)")
                self.image = self.image.transpose(Image.FLIP_LEFT_RIGHT)
            if (rot != 0.0):
                self.logger.debug("rotation: by %f degrees clockwise" % (rot))
                self.image = self.image.rotate(-rot, expand=True)

    def do_quality(self, quality):
        """Apply value of quality parameter.

        For PIL docs see
        <http://pillow.readthedocs.org/en/latest/reference/Image.html#PIL.Image.Image.convert>
        """
        if (quality == 'grey' or quality == 'gray'):
            # Checking for 1.1 gray or 20.0 grey elsewhere
            self.logger.debug("quality: converting to gray")
            self.image = self.image.convert('L')
        elif (quality == 'bitonal'):
            self.logger.debug("quality: converting to bitonal")
            self.image = self.image.convert('1')
        elif (self.image.mode not in ('1', 'L', 'RGB', 'RGBA')):
            # Need to convert from palette etc. in order to write out
            self.logger.debug("quality: converting from mode %s to RGB"
                              % (self.image.mode))
            self.image = self.image.convert('RGB')
        else:
            self.logger.debug("quality: quality (nop)")

    def do_format(self, format):
        """Apply format selection.

        Assume that for tiling applications we want jpg so return
        that unless an explicit format is requested.
        """
        fmt = ('jpg' if (format is None) else format)
        if (fmt == 'png'):
            self.logger.debug("format: png")
            self.mime_type = "image/png"
            self.output_format = fmt
            format = 'png'
        elif (fmt == 'jpg'):
            self.logger.debug("format: jpg")
            self.mime_type = "image/jpeg"
            self.output_format = fmt
            format = 'jpeg'
        elif (fmt == 'webp'):
            self.logger.debug("format: webp")
            self.mime_type = "image/webp"
            self.output_format = fmt
            format = 'webp'
        else:
            raise IIIFError(code=415, parameter='format',
                            text="Unsupported output file format (%s), only png,jpg,webp are supported." % (fmt))
        if (self.outfile is None):
            # Create temp
            f = tempfile.NamedTemporaryFile(delete=False)
            self.outfile = f.name
            self.outtmp = f.name
            self.image.save(f, format=format)
        else:
            # Save to specified location
            self.image.save(self.outfile, format=format)

    def cleanup(self):
        """Cleanup temporary output file."""
        if (self.outtmp is not None):
            try:
                os.unlink(self.outtmp)
            except OSError as e:
                self.logger.warning("Failed to cleanup tmp output file %s"
                                    % (self.outtmp))



#!/usr/bin/env python
"""Crude webserver that services IIIF Image API requests.

Relies upon IIIFManipulator objects to do any manipulations
requested and is thus very slow. Supports a number of different
versions of the specification via different base URIs (prefixes).

Simeon Warner - 2014--2016
"""

from flask import Flask, request, make_response, redirect, abort, send_file, url_for, send_from_directory

import base64
import logging
import re
import json
import optparse
import os
import os.path
from string import Template
try:  # python3
    from urllib.parse import quote as urlquote
    from urllib.request import parse_keqv_list, parse_http_list
except ImportError:  # python2
    from urllib import quote as urlquote
    from urllib2 import parse_keqv_list, parse_http_list

from iiif.error import IIIFError
from iiif.request import IIIFRequest, IIIFRequestPathError, IIIFRequestBaseURI
from iiif.info import IIIFInfo


class Config(object):
    """Class to share configuration information in IIIFHandler instances.

    Designed to allow initialization from other Config
    objects and from optparse.Values objects.
    """

    def __init__(self, *args):
        """Initialize new Config onbject copying all properties from args."""
        for arg in args:
            for k in list(arg.__dict__.keys()):
                self.__dict__[k] = arg.__dict__[k]


def no_op(self, format, *args):
    """Function that does nothing - no-op.

    Used to silence logging
    """
    pass


def html_page(title="Page Title", body=""):
    """Create HTML page as string."""
    html = "<html>\n<head><title>%s</title></head>\n<body>\n" % (title)
    html += "<h1>%s</h1>\n" % (title)
    html += body
    html += "</body>\n</html>\n"
    return html


def top_level_index_page(config):
    """HTML top-level index page."""
    http_host = request.environ.get('HTTP_HOST', '')
    title = "iiif_testserver on %s" % (http_host)
    body = "<ul>\n"
    for prefix in sorted(config.prefixes):
        body += '<li><a href="/%s">%s</a></li>\n' % (prefix, prefix)
    body += "</ul>\n"
    return html_page(title, body)


def identifiers(config):
    """Show list of identifiers for this prefix.

    Handles both the case of local file based identifiers and
    also image generators.
    """
    ids = []
    if (config.klass_name == 'gen'):
        for generator in os.listdir(config.generator_dir):
            if (generator == '__init__.py'):
                continue
            (gid, ext) = os.path.splitext(generator)
            if (ext == '.py' and
                    os.path.isfile(os.path.join(config.generator_dir, generator))):
                ids.append(gid)
    else:
        for image_file in os.listdir(config.image_dir):
            (iid, ext) = os.path.splitext(image_file)
            if (ext in ['.jpg', '.png', '.tif'] and
                    os.path.isfile(os.path.join(config.image_dir, image_file))):
                ids.append(iid)
    return ids


def prefix_index_page(config=None, prefix=None):
    """HTML index page for a specific prefix."""
    http_host = request.environ.get('HTTP_HOST', '')
    title = "Prefix %s  (from iiif_testserver on %s)" % (prefix, http_host)
    # details of this prefix handler
    body = '<p>\n'
    body += 'api_version = %s<br/>\n' % (config.api_version)
    body += 'manipulator = %s<br/>\n' % (config.klass_name)
    body += 'auth_type = %s\n</p>\n' % (config.auth_type)
    # table of identifiers and example requests
    ids = identifiers(config)
    api_version = config.api_version
    default = 'native' if api_version < '2.0' else 'default'
    body += '<table border="1">\n<tr><th align="left">Source image</th>'
    body += '<th> </th><th>full</th>'
    if (prefix != 'dummy'):
        body += '<th>256,256</th>'
        body += '<th>30deg</th>'
        if (config.include_osd):
            body += '<th> </th>'
    body += "</tr>\n"
    for identifier in sorted(ids):
        body += '<tr><th align="left">%s</th>' % (identifier)
        info = "/%s/%s/info.json" % (prefix, identifier)
        body += '<td><a href="%s">%s</a></td>' % (info, 'info')
        suffix = "full/full/0/%s" % (default)
        url = "/%s/%s/%s" % (prefix, identifier, suffix)
        body += '<td><a href="%s">%s</a></td>' % (url, suffix)
        if (prefix != 'dummy'):
            suffix = "full/256,256/0/%s" % (default)
            url = "/%s/%s/%s" % (prefix, identifier, suffix)
            body += '<td><a href="%s">%s</a></td>' % (url, suffix)
            suffix = "full/100,/30/%s" % (default)
            url = "/%s/%s/%s" % (prefix, identifier, suffix)
            body += '<td><a href="%s">%s</a></td>' % (url, suffix)
            if (config.include_osd):
                url = "/%s/%s/osd.html" % (prefix, identifier)
                body += '<td><a href="%s">OSD</a></td>' % (url)
        body += "</tr>\n"
    body += "</table<\n"
    return html_page(title, body)


def osd_page_handler(config=None, identifier=None, prefix=None, **args):
    """Produce HTML response for OpenSeadragon view of identifier."""
    template_dir = os.path.join(os.path.dirname(__file__), 'iiif', 'templates')
    with open(os.path.join(template_dir, 'testserver_osd.html'), 'r') as f:
        template = f.read()
    d = dict(prefix=prefix,
             identifier=identifier,
             api_version=config.api_version,
             osd_version='2.0.0',
             osd_uri='/openseadragon200/openseadragon.min.js',
             osd_images_prefix='/openseadragon200/images',
             osd_height=500,
             osd_width=500,
             info_json_uri='info.json')
    return make_response(Template(template).safe_substitute(d))


def host_port_prefix(host, port, prefix):
    """Return URI composed of scheme, server, port, and prefix."""
    uri = "http://" + host
    if (host != 80):
        uri += ':' + str(port)
    if (prefix):
        uri += '/' + prefix
    return uri


class IIIFHandler(object):
    """IIIFHandler class."""

    def __init__(self, prefix, identifier, config, klass, auth):
        """Initialize IIIFHandler setting key configurations.

        Positional parameters:
        prefix -- URI prefix (without leading or trailing slashes)
        identifier -- identifier of image
        config -- instance of Config class
        klass -- IIIFManipulator sub-class to do manipulations
        auth -- IIIFAuth sub-class for auth
        """
        self.prefix = prefix
        self.identifier = identifier
        self.config = config
        self.klass = klass
        self.api_version = config.api_version
        self.auth = auth
        self.degraded = False
        self.logger = logging.getLogger('IIIFHandler')
        #
        # Create objects to process request
        self.iiif = IIIFRequest(api_version=self.api_version,
                                identifier=self.identifier)
        self.manipulator = klass(api_version=self.api_version)
        #
        # Set up auth object with locations if not already done
        if (self.auth and not self.auth.login_uri):
            self.auth.login_uri = self.server_and_prefix + '/login'
            self.auth.logout_uri = self.server_and_prefix + '/logout'
            self.auth.access_token_uri = self.server_and_prefix + '/token'
        #
        # Response headers
        # -- All responses should have CORS header
        self.headers = {'Access-control-allow-origin': '*'}

    @property
    def server_and_prefix(self):
        """Server and prefix from config."""
        return(host_port_prefix(self.config.host, self.config.port, self.prefix))

    @property
    def json_mime_type(self):
        """Return the MIME type for a JSON response.

        For version 2.0+ the server must return json-ld MIME type if that
        format is requested. Implement for 1.1 also.
        http://iiif.io/api/image/2.1/#information-request
        """
        mime_type = "application/json"
        if (self.api_version >= '1.1' and
                'Accept' in request.headers):
            mime_type = do_conneg(request.headers['Accept'], [
                                  'application/ld+json']) or mime_type
        return mime_type

    @property
    def file(self):
        """Filename property for the source image for the current identifier."""
        file = None
        if (self.config.klass_name == 'gen'):
            for ext in ['.py']:
                file = os.path.join(
                    self.config.generator_dir, self.identifier + ext)
                if (os.path.isfile(file)):
                    return file
        else:
            for ext in ['.jpg', '.png', '.tif']:
                file = os.path.join(self.config.image_dir,
                                    self.identifier + ext)
                if (os.path.isfile(file)):
                    return file
        # failed, show list of available identifiers as error
        available = "\n ".join(identifiers(self.config))
        raise IIIFError(code=404, parameter="identifier",
                        text="Image resource '" + self.identifier + "' not found. Local resources available:" + available + "\n")

    def add_compliance_header(self):
        """Add IIIF Compliance level header to response."""
        if (self.manipulator.compliance_uri is not None):
            self.headers['Link'] = '<' + \
                self.manipulator.compliance_uri + '>;rel="profile"'

    def make_response(self, content, code=200, headers=None):
        """Wrapper around Flask.make_response which also adds any local headers."""
        if headers:
            for header in headers:
                self.headers[header] = headers[header]
        return make_response(content, code, self.headers)

    def image_information_response(self):
        """Parse image information request and create response."""
        dr = degraded_request(self.identifier)
        if (dr):
            self.logger.info("image_information: degraded %s -> %s" %
                             (self.identifier, dr))
            self.degraded = self.identifier
            self.identifier = dr
        else:
            self.logger.info("image_information: %s" % (self.identifier))
        # get size
        self.manipulator.srcfile = self.file
        self.manipulator.do_first()
        # most of info.json comes from config, a few things specific to image
        info = {'tile_height': self.config.tile_height,
                'tile_width': self.config.tile_width,
                'scale_factors': self.config.scale_factors
                }
        # calculate scale factors if not hard-coded
        if ('auto' in self.config.scale_factors):
            info['scale_factors'] = self.manipulator.scale_factors(
                self.config.tile_width, self.config.tile_height)
        i = IIIFInfo(conf=info, api_version=self.api_version)
        i.server_and_prefix = self.server_and_prefix
        i.identifier = self.iiif.identifier
        i.width = self.manipulator.width
        i.height = self.manipulator.height
        if (self.api_version >= '2.0'):
            # FIXME - should come from manipulator
            i.qualities = ["default", "color", "gray"]
        else:
            # FIXME - should come from manipulator
            i.qualities = ["native", "color", "gray"]
        i.formats = ["jpg", "png"]  # FIXME - should come from manipulator
        if (self.auth):
            self.auth.add_services(i)
        return self.make_response(i.as_json(),
                                  headers={"Content-Type": self.json_mime_type})

    def image_request_response(self, path):
        """Parse image request and create response."""
        # Parse the request in path
        if (len(path) > 1024):
            raise IIIFError(code=414,
                            text="URI Too Long: Max 1024 chars, got %d\n" % len(path))
        # print "GET " + path
        try:
            self.iiif.identifier = self.identifier
            self.iiif.parse_url(path)
        except IIIFRequestPathError as e:
            # Reraise as IIIFError with code=404 because we can't tell
            # whether there was an encoded slash in the identifier or
            # whether there was a bad number of path segments.
            raise IIIFError(code=404, text=e.text)
        except IIIFRequestBaseURI as e:
            info_uri = self.server_and_prefix + '/' + \
                urlquote(self.iiif.identifier) + '/info.json'
            raise IIIFError(code=303,
                            headers={'Location': info_uri})
        except IIIFError as e:
            # Pass through
            raise e
        except Exception as e:
            # Something completely unexpected => 500
            raise IIIFError(code=500,
                            text="Internal Server Error: unexpected exception parsing request (" + str(e) + ")")
        dr = degraded_request(self.identifier)
        if (dr):
            self.logger.info("image_request: degraded %s -> %s" %
                             (self.identifier, dr))
            self.degraded = self.identifier
            self.identifier = dr
            self.iiif.quality = 'gray'
        else:
            # Parsed request OK, attempt to fulfill
            self.logger.info("image_request: %s" % (self.identifier))
        file = self.file
        self.manipulator.srcfile = file
        self.manipulator.do_first()
        if (self.api_version < '2.0' and
                self.iiif.format is None and
                'Accept' in request.headers):
            # In 1.0 and 1.1 conneg was specified as an alternative to format, see:
            # http://iiif.io/api/image/1.0/#format
            # http://iiif.io/api/image/1.1/#parameters-format
            formats = {'image/jpeg': 'jpg', 'image/tiff': 'tif',
                       'image/png': 'png', 'image/gif': 'gif',
                       'image/jp2': 'jps', 'application/pdf': 'pdf'}
            accept = do_conneg(request.headers['Accept'], list(formats.keys()))
            # Ignore Accept header if not recognized, should this be an error
            # instead?
            if (accept in formats):
                self.iiif.format = formats[accept]
        (outfile, mime_type) = self.manipulator.derive(file, self.iiif)
        # FIXME - find efficient way to serve file with headers
        self.add_compliance_header()
        return send_file(open(outfile, 'rb'), mimetype=mime_type)

    def error_response(self, e):
        """Make response for an IIIFError e.

        Also add compliance header.
        """
        self.add_compliance_header()
        return self.make_response(*e.image_server_response(self.api_version))


def iiif_info_handler(prefix=None, identifier=None, config=None, klass=None, auth=None, **args):
    """Handler for IIIF Image Information requests."""
    if (not auth or degraded_request(identifier) or auth.info_authz()):
        # go ahead with request as made
        if (auth):
            print("Authorized for image %s" % identifier)
        i = IIIFHandler(prefix, identifier, config, klass, auth)
        try:
            return i.image_information_response()
        except IIIFError as e:
            return i.error_response(e)
    elif (auth.info_authn()):
        # authn but not authz -> 401
        abort(401)
    else:
        # redirect to degraded
        response = redirect(host_port_prefix(
            config.host, config.port, prefix) + '/' + identifier + '-deg/info.json')
        response.headers['Access-control-allow-origin'] = '*'
        return response
iiif_info_handler.provide_automatic_options = False


def iiif_image_handler(prefix=None, identifier=None, path=None, config=None, klass=None, auth=None, **args):
    """Handler for IIIF Image Requests.

    Behaviour for case of a non-authn or non-authz case is to
    return 403.
    """
    if (not auth or degraded_request(identifier) or auth.image_authz()):
        # serve image
        if (auth):
            print("Authorized for image %s" % identifier)
        i = IIIFHandler(prefix, identifier, config, klass, auth)
        try:
            return i.image_request_response(path)
        except IIIFError as e:
            return i.error_response(e)
    else:
        # redirect to degraded (for not authz and for authn but not authz too)
        degraded_uri = host_port_prefix(
            config.host, config.port, prefix) + '/' + identifier + '-deg/' + path
        print("Redirection to degraded: %s" % degraded_uri)
        response = redirect(degraded_uri)
        response.headers['Access-control-allow-origin'] = '*'
        return response
iiif_image_handler.provide_automatic_options = False


def degraded_request(identifier):
    """Return True (non-degraded id) if this is a degraded request, False otherwise."""
    if identifier.endswith('-deg'):
        return identifier[:-4]
    return False


def options_handler(**args):
    """Handler to respond to OPTIONS requests."""
    headers = {'Access-Control-Allow-Origin': '*',
               'Access-Control-Allow-Methods': 'GET,OPTIONS',
               'Access-Control-Allow-Headers': 'Origin, Accept, Authorization'}
    return make_response("", 200, headers)


def parse_accept_header(accept):
    """Parse an HTTP Accept header.

    Parses *accept*, returning a list with pairs of
    (media_type, q_value), ordered by q values.

    Adapted from <https://djangosnippets.org/snippets/1042/>
    """
    result = []
    for media_range in accept.split(","):
        parts = media_range.split(";")
        media_type = parts.pop(0).strip()
        media_params = []
        q = 1.0
        for part in parts:
            (key, value) = part.lstrip().split("=", 1)
            if key == "q":
                q = float(value)
            else:
                media_params.append((key, value))
        result.append((media_type, tuple(media_params), q))
    result.sort(key=lambda x: -x[2])
    return result


def parse_authorization_header(value):
    """Parse the Authenticate header.

    Returns nothing on failure, opts hash on success with type='basic' or 'digest'
    and other params.

    <http://nullege.com/codes/search/werkzeug.http.parse_authorization_header>
    <http://stackoverflow.com/questions/1349367/parse-an-http-request-authorization-header-with-python>
    <http://bugs.python.org/file34041/0001-Add-an-authorization-header-to-the-initial-request.patch>
    """
    try:
        (auth_type, auth_info) = value.split(' ', 1)
        auth_type = auth_type.lower()
    except ValueError as e:
        return
    if (auth_type == 'basic'):
        try:
            decoded = base64.b64decode(auth_info).decode(
                'utf-8')  # b64decode gives bytes in python3
            (username, password) = decoded.split(':', 1)
        except ValueError:  # Exception as e:
            return
        return {'type': 'basic', 'username': username, 'password': password}
    elif (auth_type == 'digest'):
        auth_map = parse_keqv_list(parse_http_list(auth_info))
        print(auth_map)
        for key in 'username', 'realm', 'nonce', 'uri', 'response':
            if key not in auth_map:
                return
            if 'qop' in auth_map:
                if not auth_map.get('nc') or not auth_map.get('cnonce'):
                    return
        auth_map['type'] = 'digest'
        return auth_map
    else:
        # unknown auth type
        return


def do_conneg(accept, supported):
    """Parse accept header and look for preferred type in supported list.

    accept parameter is HTTP header, supported is a list of MIME types
    supported by the server. Returns the supported MIME type with highest
    q value in request, else None.
    """
    for result in parse_accept_header(accept):
        mime_type = result[0]
        if (mime_type in supported):
            return(mime_type)
    return(None)

######################################################################


def setup_auth_paths(app, auth, prefix, params):
    """Add URL rules for auth paths."""
    base = '/' + prefix + '/'
    app.add_url_rule(base + 'login', prefix + 'login_handler',
                     auth.login_handler, defaults=params)
    app.add_url_rule(base + 'logout', prefix + 'logout_handler',
                     auth.logout_handler, defaults=params)
    if (auth.client_id_handler):
        app.add_url_rule(base + 'client', prefix + 'client_id_handler',
                         auth.client_id_handler, defaults=params)
    app.add_url_rule(base + 'token', prefix + 'access_token_handler',
                     auth.access_token_handler, defaults=params)
    if (auth.home_handler):
        app.add_url_rule(base + 'home', prefix + 'home_handler',
                         auth.home_handler, defaults=params)


def make_prefix(api_version, manipulator, auth_type):
    """Make prefix string based on configuration parameters."""
    prefix = "%s_%s" % (api_version, manipulator)
    if (auth_type and auth_type != 'none'):
        prefix += '_' + auth_type
    return(prefix)


def split_option(comma_sep_str):
    """Split a comma separated option."""
    return comma_sep_str.split(',')  # FIXME - make more flexible


def setup_options():
    """Parse options and arguments."""
    p = optparse.OptionParser(description='IIIF Image Testserver')
    p.add_option('--host', default='localhost',
                 help="Server host (default %default)")
    p.add_option('--port', '-p', type='int', default=8000,
                 help="Server port (default %default)")
    p.add_option('--image-dir', '-d', default='testimages',
                 help="Image directory (default %default)")
    p.add_option('--generator-dir', default='iiif/generators',
                 help="Generator directory for manipulator='gen' (default %default)")
    p.add_option('--tile-height', type='int', default=256,
                 help="Tile height (default %default)")
    p.add_option('--tile-width', type='int', default=256,
                 help="Tile width (default %default)")
    p.add_option('--scale-factors', default='auto',
                 help="Set of tile scale factors or 'auto' to calculate for each image "
                      "such that there are tiles up to the full image (default %default)")
    p.add_option('--api-versions', default='1.0,1.1,2.0,2.1',
                 help="Set of API versions to support (default %default)")
    p.add_option('--manipulators', default='pil',
                 help="Set of manipuators to instantiate. May be dummy,netpbm,pil "
                      "or gen for generated image. (default %default")
    p.add_option('--auth-types', default='none',
                 help="Set of authentication types to support (default %default)")
    p.add_option('--gauth-client-secret', default='client_secret.json',
                 help="Name of file with Google auth client secret (default %default)")
    p.add_option('--pages-dir', default='testpages',
                 help="Test pages directory (default %default)")
    p.add_option('--include-osd', action='store_true',
                 help="Include a page with OpenSeadragon for each source")
    p.add_option('--draft', action='store_true',
                 help="Enable features implementing draft of IIIF auth specification")
    p.add_option('--debug', action='store_true',
                 help="Set debug mode for web application. INSECURE!")
    p.add_option('--verbose', '-v', action='store_true',
                 help="Be verbose")
    p.add_option('--quiet', '-q', action='store_true',
                 help="Minimal output only")
    (opt, args) = p.parse_args()

    if (opt.debug):
        opt.verbose = True
    elif (opt.verbose):
        opt.quiet = False

    # Split list arguments
    opt.scale_factors = split_option(opt.scale_factors)
    opt.manipulators = split_option(opt.manipulators)
    opt.api_versions = split_option(opt.api_versions)
    opt.auth_types = split_option(opt.auth_types)

    # Draft features...
    if (opt.draft and 'gauth' not in opt.auth_types):
        opt.auth_types.append('gauth')
    if (opt.draft and 'basic' not in opt.auth_types):
        opt.auth_types.append('basic')

    return(opt)


def add_handler(app, config, prefixes):
    """Add a single handler to the app.

    Adds handlers to app, with config from config. Add prefix to list
    in prefixes.
    """
    wsgi_prefix = make_prefix(
        config.api_version, config.klass_name, config.auth_type)
    prefix = wsgi_prefix
    if (config.container_prefix):
        prefix = os.path.join(config.container_prefix, wsgi_prefix)
    prefixes.append(prefix)
    auth = None
    if (config.auth_type is None or config.auth_type == 'none'):
        pass
    elif (config.auth_type == 'gauth'):
        from iiif.auth_google import IIIFAuthGoogle
        auth = IIIFAuthGoogle(
            client_secret_file=config.gauth_client_secret_file)
    elif (config.auth_type == 'basic'):
        from iiif.auth_basic import IIIFAuthBasic
        auth = IIIFAuthBasic()
    else:
        print("Unknown auth type %s, ignoring" % (config.auth_type))
        return
    klass = None
    if (config.klass_name == 'pil'):
        from iiif.manipulator_pil import IIIFManipulatorPIL
        klass = IIIFManipulatorPIL
    elif (config.klass_name == 'netpbm'):
        from iiif.manipulator_netpbm import IIIFManipulatorNetpbm
        klass = IIIFManipulatorNetpbm
    elif (config.klass_name == 'dummy'):
        from iiif.manipulator import IIIFManipulator
        klass = IIIFManipulator
    elif (config.klass_name == 'gen'):
        from iiif.manipulator_gen import IIIFManipulatorGen
        klass = IIIFManipulatorGen
    else:
        print("Unknown manipulator type %s, ignoring" % (config.klass_name))
        return
    print("Installing %s IIIFManipulator at /%s/ v%s %s" %
          (config.klass_name, prefix, config.api_version, config.auth_type))
    params = dict(config=config, klass=klass, auth=auth, prefix=prefix)
    app.add_url_rule('/' + wsgi_prefix, 'prefix_index_page',
                     prefix_index_page, defaults={'config': config, 'prefix': prefix})
    app.add_url_rule('/' + wsgi_prefix + '/<string(minlength=1):identifier>/info.json',
                     'options_handler', options_handler, methods=['OPTIONS'])
    app.add_url_rule('/' + wsgi_prefix + '/<string(minlength=1):identifier>/info.json',
                     'iiif_info_handler', iiif_info_handler, methods=['GET'], defaults=params)
    if (config.include_osd):
        app.add_url_rule('/' + wsgi_prefix + '/<string(minlength=1):identifier>/osd.html',
                         'osd_page_handler', osd_page_handler, methods=['GET'], defaults=params)
    app.add_url_rule('/' + wsgi_prefix + '/<string(minlength=1):identifier>/<path:path>',
                     'iiif_image_handler', iiif_image_handler, methods=['GET'], defaults=params)
    if (auth):
        setup_auth_paths(app, auth, wsgi_prefix, params)
    # redirects to info.json must come after auth
    app.add_url_rule('/' + wsgi_prefix + '/<string(minlength=1):identifier>',
                     'iiif_info_handler', redirect_to='/' + prefix + '/<identifier>/info.json')
    app.add_url_rule('/' + wsgi_prefix + '/<string(minlength=1):identifier>/',
                     'iiif_info_handler', redirect_to='/' + prefix + '/<identifier>/info.json')


def serve_static(filename=None, prefix=None, basedir=''):
    """Handler for static files under basedir."""
    return send_from_directory(os.path.join('third_party', prefix),
                               filename)


def create_app(opt):
    """Create Flask application with one or more IIIF handlers."""
    logging_level = logging.WARNING
    if (opt.verbose):
        logging_level = logging.INFO
    elif (opt.quiet):
        logging_level = logging.ERROR
    logging.basicConfig(format='%(name)s: %(message)s', level=logging_level)

    # Create Flask app
    app = Flask(__name__, static_url_path='/' + opt.pages_dir)
    Flask.secret_key = "SECRET_HERE"
    app.debug = opt.debug

    # Create shared configuration dict based on options
    config = Config(opt)
    config.homedir = os.path.dirname(os.path.realpath(__file__))
    config.gauth_client_secret_file = os.path.join(
        config.homedir, config.gauth_client_secret)

    prefixes = []
    for api_version in opt.api_versions:
        for klass_name in opt.manipulators:
            for auth_type in opt.auth_types:
                # auth only for >=2.1
                if (auth_type != 'none' and float(api_version) < 2.1):
                    continue
                handler_config = Config(config)
                handler_config.api_version = api_version
                handler_config.klass_name = klass_name
                handler_config.auth_type = auth_type
                add_handler(app, handler_config, prefixes)

    # Index page
    config.prefixes = prefixes
    app.add_url_rule('/', 'top_level_index_page',
                     top_level_index_page, defaults={'config': config})
    if (config.include_osd):
        # OpenSeadragon files
        # app.add_url_rule('/openseadragon100/<path:filename>', 'OSD pages', serve_static, defaults={'prefix':'openseadragon100','basedir':'third_party'})
        # app.add_url_rule('/openseadragon121/<path:filename>', 'OSD pages', serve_static, defaults={'prefix':'openseadragon121','basedir':'third_party'})
        app.add_url_rule('/openseadragon200/<path:filename>', 'OSD pages', serve_static,
                         defaults={'prefix': 'openseadragon200', 'basedir': 'third_party'})

    return(app)


if __name__ == '__main__':
    # Command line, run own server
    pidfile = os.path.basename(__file__)[:-3] + '.pid'  # strip .py, add .pid
    with open(pidfile, 'w') as fh:
        fh.write("%d\n" % os.getpid())
        fh.close()
    opt = setup_options()
    opt.container_prefix = ''
    app = create_app(opt)
    print("Starting test server on http://%s:%d/ ..." % (opt.host, opt.port))
    app.run(host=opt.host, port=opt.port)
else:
    opt = optparse.Values()
    opt.verbose = 1
    opt.debug = 1
    opt.draft = 1
    mydir = os.path.dirname(os.path.realpath(__file__))
    opt.pages_dir = mydir + '/testpages'
    opt.image_dir = mydir + '/testimages'
    opt.generator_dir = mydir + '/iiif/generators'
    opt.tile_width = 512
    opt.tile_height = 512
    opt.scale_factors = ['auto']
    opt.api_versions = ['1.0', '1.1', '2.0', '2.1']
    opt.auth_types = ['none', 'gauth', 'basic']
    opt.manipulators = ['dummy', 'netpbm', 'pil']
    opt.include_osd = False
    opt.container_prefix = 'iiif_auth_test'
    # Should get the following from WSGI environ,
    # see https://code.google.com/p/modwsgi/wiki/ConfigurationGuidelines
    opt.host = 'resync.library.cornell.edu'  # FIXME - get from WSGI
    opt.port = 80
    opt.gauth_client_secret = 'client_secret.json'
    app = create_app(opt)



"""Create and parse IIIF request URLs.

This class is a thorough implementation of restrictions in the
IIIF specification. It does not add any implementation specific
restrictions.
"""
import re
import string
try:  # python3
    from urllib.parse import quote as urlquote
    from urllib.parse import unquote as urlunquote
except ImportError:  # python2
    from urllib import quote as urlquote
    from urllib import unquote as urlunquote

from .error import IIIFError, IIIFZeroSizeError


class IIIFRequestError(IIIFError):
    """Subclass of IIIFError for request parsing errors."""

    def __init__(self, code=400, parameter='unknown',
                 text='Request parsing error'):
        """Initialize."""
        super(IIIFRequestError, self).__init__(
            code=code, parameter=parameter, text=text)


class IIIFRequestPathError(IIIFRequestError):
    """Subclass of IIIFRequestError for path size errors.

    WSGI and other applications where we can't tell whether
    there was an encoded slash or an unencoded slash in an
    identifier or (elsewhere) may wish to trap this particular
    error and return a 404 Not Found instead of 400.
    """

    pass


class IIIFRequestBaseURI(Exception):
    """Subclass of Exception to indicate request for base URI."""

    pass


class IIIFRequest(object):
    """Implement IIIF request URL syntax.

    There are two URL forms defined in section 2:
    http[s]://server/[prefix/]identifier/region/size/rotation/quality[.format]
    or
    http[s]://server/[prefix/]identifier/info.format

    We also detect and throw a special IIIFRequestBaseURI exception then
    the naked base URI is requested:
    http[s]://server/[prefix/]identifier(/?)

    The attribites of objects of this class follow the names here except that
    baseurl is used for "http[s]://server/[prefix/]". If baseurl is not set
    then a relative URL will be created or parsed.
    """

    def __init__(self, api_version='2.1', baseurl='',
                 allow_slashes_in_identifier=False, **params):
        """Initialize Request object and optionally set attributes.

        Any of the attributes of the Request object may be set via
        named parameters.

        Current API version assumed if not specified. If another API
        version is to be used then this should be set on creation via the
        api_version parameter.

        Unless specified the baseurl will be set to nothing ("").
        """
        self.clear()
        self.baseurl = baseurl
        self.api_version = api_version
        self.allow_slashes_in_identifier = allow_slashes_in_identifier
        self._setattrs(**params)

    def clear(self):
        """Clear all data that might pertain to an individual IIIF URL.

        Does not change/reset the baseurl or API version which might be
        useful in a sequence of calls.
        """
        # API parameters
        self.identifier = None
        self.region = None
        self.size = None
        self.rotation = None
        self.quality = None
        self.format = None
        self.info = None
        # Derived data and flags
        self.region_full = False
        self.region_square = False
        self.region_pct = False
        self.region_xywh = None  # (x,y,w,h)
        self.size_full = False
        self.size_max = False  # new in 2.1
        self.size_pct = None
        self.size_bang = None
        self.size_wh = None     # (w,h)
        self.rotation_mirror = False
        self.rotation_deg = 0.0

    @property
    def api_version(self):
        """Get api_version value."""
        return self._api_version

    @api_version.setter
    def api_version(self, v):
        """Set the api_version and associated configurations."""
        self._api_version = v
        if (self._api_version >= '2.0'):
            self.default_quality = 'default'
            self.allowed_qualities = ['default', 'color', 'bitonal', 'gray']
        else:  # versions 1.0 and 1.1
            self.default_quality = 'native'
            self.allowed_qualities = ['native', 'color', 'bitonal', 'grey']

    def _setattrs(self, **params):
        # Set one of the allowed request parameters.
        #
        # Will silently ignore any unknown parameters.
        for k in ('identifier', 'region', 'size',
                  'rotation', 'quality', 'format', 'info'):
            if (k in params):
                setattr(self, k, params[k])

    def quote(self, path_segment):
        """Quote parameters in IIIF URLs.

        Quote by percent-encoding "%" and gen-delims of RFC3986 exept colon

        http://tools.ietf.org/html/rfc3986#section-2.3
        gen-delims  = ":" / "/" / "?" / "#" / "[" / "]" / "@"

        to-encode = "%" / "/" / "?" / "#" / "[" / "]" / "@"
        """
        # FIXME - quotes too much
        return(urlquote(path_segment, "-._~!$&'()*+,;=:"))

    def url(self, **params):
        """Build a URL path for image or info request.

        An IIIF Image request with parameterized form is assumed unless
        the info parameter is specified, in which case an Image Information
        request URI is constructred.
        """
        self._setattrs(**params)
        path = self.baseurl + self.quote(self.identifier) + "/"
        if (self.info):
            # info request
            path += "info"
            format = self.format if self.format else "json"
        else:
            # region
            if self.region:
                region = self.region
            elif self.region_xywh:
                region = "%d,%d,%d,%d" % tuple(self.region_xywh)
            else:
                region = "full"
            # size
            if self.size:
                size = self.size
            elif self.size_wh:
                if (self.size_wh[0] is None):
                    size = ",%d" % (self.size_wh[1])
                elif (self.size_wh[1] is None):
                    size = "%d," % (self.size_wh[0])
                else:
                    size = "%d,%d" % (self.size_wh[0], self.size_wh[1])
            elif (self.size_max and self.api_version >= '2.1'):
                size = 'max'
            else:
                size = "full"
            # rotation and quality
            rotation = self.rotation if self.rotation else "0"
            quality = self.quality if self.quality else self.default_quality
            # parameterized form
            path += self.quote(region) + "/" +\
                self.quote(size) + "/" +\
                self.quote(rotation) + "/" +\
                self.quote(quality)
            format = self.format
        if (format):
            path += "." + format
        return(path)

    def parse_url(self, url):
        """Parse an IIIF API URL path and each component.

        Will parse a URL or URL path that accords with either the
        parametrized or info request forms. Will raise an
        IIIFRequestError on failure. A wrapper for the split_url()
        and parse_parameters() methods.

        Note that behavior of split_url() depends on whether
        self.identifier is set.
        """
        self.split_url(url)
        if (not self.info):
            self.parse_parameters()
        return(self)

    def _allow_slashes_in_identifier_munger(self, segs):
        # Big fudge to account for the fact that WSGI doesn't
        # preserve escaping when sending paths to application.
        # Thus we do very simply pattern matching to assume extra
        # slashes are part of the identifier.
        if (len(segs) > 2 and re.match(r'''info\.\w+$''', segs[-1])):
            # Looks like an information request, combine all
            # but last segment into first/identifier
            segs = ['/'.join(segs[0:-1]), segs[-1]]
        elif (len(segs) > 5):
            # Assume image request, combine front segments
            # into first/identifier leaving 5 total
            segs = ['/'.join(segs[0:-4]),
                    segs[-4], segs[-3], segs[-2], segs[-1]]
        return segs

    def split_url(self, url):
        """Parse an IIIF API URL path into components.

        Will parse a URL or URL path that accords with either the
        parametrized or info API forms. Will raise an IIIFRequestError on
        failure.

        If self.identifier is set then url is assumed not to include the
        identifier.
        """
        # clear data first
        identifier = self.identifier
        self.clear()
        # url must start with baseurl if set (including slash)
        if (self.baseurl is not None):
            (path, num) = re.subn('^' + self.baseurl, '', url, 1)
            if (num != 1):
                raise IIIFRequestError(
                    text="Request URL does not start with base URL")
            url = path
        # Break up by path segments, count to decide format
        segs = url.split('/')
        if (identifier is not None):
            segs.insert(0, identifier)
        elif (self.allow_slashes_in_identifier):
            segs = self._allow_slashes_in_identifier_munger(segs)
        # Now have segments with identifier as first
        if (len(segs) > 5):
            raise IIIFRequestPathError(
                text="Request URL (%s) has too many path segments" % url)
        elif (len(segs) == 5):
            self.identifier = urlunquote(segs[0])
            self.region = urlunquote(segs[1])
            self.size = urlunquote(segs[2])
            self.rotation = urlunquote(segs[3])
            self.quality = self.strip_format(urlunquote(segs[4]))
            self.info = False
        elif (len(segs) == 2):
            self.identifier = urlunquote(segs[0])
            info_name = self.strip_format(urlunquote(segs[1]))
            if (info_name != "info"):
                raise IIIFRequestError(
                    text="Bad name for Image Information")
            if (self.api_version == '1.0'):
                if (self.format not in ['json', 'xml']):
                    raise IIIFRequestError(
                        text="Invalid format for Image Information (json and xml allowed)")
            elif (self.format != 'json'):
                raise IIIFRequestError(
                    text="Invalid format for Image Information (only json allowed)")
            self.info = True
        elif (len(segs) == 1):
            self.identifier = urlunquote(segs[0])
            raise IIIFRequestBaseURI()
        else:
            raise IIIFRequestPathError(
                text="Bad number of path segments in request")
        return(self)

    def strip_format(self, str_and_format):
        """Look for optional .fmt at end of URI.

        The format must start with letter. Note that we want to catch
        the case of a dot and no format (format='') which is different
        from no dot (format=None)

        Sets self.format as side effect, returns possibly modified string
        """
        m = re.match("(.+)\.([a-zA-Z]\w*)$", str_and_format)
        if (m):
            # There is a format string at end, chop off and store
            str_and_format = m.group(1)
            self.format = (m.group(2) if (m.group(2) is not None) else '')
        return(str_and_format)

    def parse_parameters(self):
        """Parse the parameters of an Image Information request.

        Will throw an IIIFRequestError on failure, set attributes on
        success. Care is taken not to change any of the artibutes
        which store path components. All parsed values are stored
        in new attributes.
        """
        self.parse_region()
        self.parse_size()
        self.parse_rotation()
        self.parse_quality()
        self.parse_format()

    def parse_region(self):
        """Parse the region component of the path.

        /full/ -> self.region_full = True (test this first)
        /square/ -> self.region_square = True (test this second)
        /x,y,w,h/ -> self.region_xywh = (x,y,w,h)
        /pct:x,y,w,h/ -> self.region_xywh and self.region_pct = True

        Will throw errors if the parameters are illegal according to the
        specification but does not know about and thus cannot do any tests
        against any image being manipulated.
        """
        self.region_full = False
        self.region_square = False
        self.region_pct = False
        if (self.region is None or self.region == 'full'):
            self.region_full = True
            return
        if (self.api_version >= '2.1' and self.region == 'square'):
            self.region_square = True
            return
        xywh = self.region
        pct_match = re.match('pct:(.*)$', self.region)
        if (pct_match):
            xywh = pct_match.group(1)
            self.region_pct = True
        # Now whether this was pct: or now, we expect 4 values...
        str_values = xywh.split(',', 5)
        if (len(str_values) != 4):
            raise IIIFRequestError(
                code=400, parameter="region",
                text="Bad number of values in region specification, "
                     "must be x,y,w,h but got %d value(s) from '%s'" %
                     (len(str_values), xywh))
        values = []
        for str_value in str_values:
            # Must be either integer (not pct) or interger/float (pct)
            if (pct_match):
                try:
                    # This is rather more permissive that the iiif spec
                    value = float(str_value)
                except ValueError:
                    raise IIIFRequestError(
                        parameter="region",
                        text="Bad floating point value for percentage in "
                             "region (%s)." % str_value)
                if (value > 100.0):
                    raise IIIFRequestError(
                        parameter="region",
                        text="Percentage over value over 100.0 in region "
                             "(%s)." % str_value)
            else:
                try:
                    value = int(str_value)
                except ValueError:
                    raise IIIFRequestError(
                        parameter="region",
                        text="Bad integer value in region (%s)." % str_value)
            if (value < 0):
                raise IIIFRequestError(
                    parameter="region",
                    text="Negative values not allowed in region (%s)." %
                    str_value)
            values.append(value)
        # Zero size region is w or h are zero (careful that they may be float)
        if (values[2] == 0.0 or values[3] == 0.0):
            raise IIIFZeroSizeError(
                code=400, parameter="region",
                text="Zero size region specified (%s))." % xywh)
        self.region_xywh = values

    def parse_size(self, size=None):
        """Parse the size component of the path.

        /full/ -> self.size_full = True
        /max/ -> self.size_mac = True  (2.1 and up)
        /w,/ -> self.size_wh = (w,None)
        /,h/ -> self.size_wh = (None,h)
        /w,h/ -> self.size_wh = (w,h)
        /pct:p/ -> self.size_pct = p
        /!w,h/ -> self.size_wh = (w,h), self.size_bang = True

        Expected use:
          (w,h) = iiif.size_to_apply(region_w,region_h)
          if (q is None):
              # full image
          else:
              # scale to w by h
        Returns (None,None) if no scaling is required.
        """
        if (size is not None):
            self.size = size
        self.size_pct = None
        self.size_bang = False
        self.size_full = False
        self.size_wh = (None, None)
        if (self.size is None or self.size == 'full'):
            self.size_full = True
            return
        elif (self.size == 'max' and self.api_version >= '2.1'):
            self.size_max = True
            return
        pct_match = re.match('pct:(.*)$', self.size)
        if (pct_match is not None):
            pct_str = pct_match.group(1)
            try:
                self.size_pct = float(pct_str)
            except ValueError:
                raise IIIFRequestError(
                    code=400, parameter="size",
                    text="Percentage size value must be a number, got "
                         "'%s'." % (pct_str))
            # Note that Image API specificaton places no upper limit on
            # size so none is implemented here.
            if (self.size_pct < 0.0):
                raise IIIFRequestError(
                    code=400, parameter="size",
                    text="Base size percentage, must be > 0.0, got %f." %
                    (self.size_pct))
        else:
            if (self.size[0] == '!'):
                # Have "!w,h" form
                size_no_bang = self.size[1:]
                (mw, mh) = self._parse_w_comma_h(size_no_bang, 'size')
                if (mw is None or mh is None):
                    raise IIIFRequestError(
                        code=400, parameter="size",
                        text="Illegal size requested: both w,h must be "
                             "specified in !w,h requests.")
                self.size_wh = (mw, mh)
                self.size_bang = True
            else:
                # Must now be "w,h", "w," or ",h"
                self.size_wh = self._parse_w_comma_h(self.size, 'size')
            # Sanity check w,h
            (w, h) = self.size_wh
            if ((w is not None and w <= 0) or
                    (h is not None and h <= 0)):
                raise IIIFZeroSizeError(
                    code=400, parameter='size',
                    text="Size parameters request zero size result image.")

    def _parse_w_comma_h(self, whstr, param):
        """Utility to parse "w,h" "w," or ",h" values.

        Returns (w,h) where w,h are either None or ineteger. Will
        throw a ValueError if there is a problem with one or both.
        """
        try:
            (wstr, hstr) = whstr.split(',', 2)
            w = self._parse_non_negative_int(wstr, 'w')
            h = self._parse_non_negative_int(hstr, 'h')
        except ValueError as e:
            raise IIIFRequestError(
                code=400, parameter=param,
                text="Illegal %s value (%s)." % (param, str(e)))
        if (w is None and h is None):
            raise IIIFRequestError(
                code=400, parameter=param,
                text="Must specify at least one of w,h for %s." % (param))
        return(w, h)

    def _parse_non_negative_int(self, istr, name):
        """Parse integer from string (istr).

        The (name) parameter is used just for IIIFRequestError message
        generation to indicate what the error is in.
        """
        if (istr == ''):
            return(None)
        try:
            i = int(istr)
        except ValueError:
            raise ValueError("Failed to extract integer value for %s" % (name))
        if (i < 0):
            raise ValueError("Illegal negative value for %s" % (name))
        return(i)

    def parse_rotation(self, rotation=None):
        """Check and interpret rotation.

        Uses value of self.rotation as starting point unless rotation
        parameter is specified in the call. Sets self.rotation_deg to a
        floating point number 0 <= angle < 360. Includes translation of
        360 to 0. If there is a prefix bang (!) then self.rotation_mirror
        will be set True, otherwise it will be False.
        """
        if (rotation is not None):
            self.rotation = rotation
        self.rotation_deg = 0.0
        self.rotation_mirror = False
        if (self.rotation is None):
            return
        # Look for ! prefix first
        if (self.rotation[0] == '!'):
            self.rotation_mirror = True
            self.rotation = self.rotation[1:]
        # Interpret value now
        try:
            self.rotation_deg = float(self.rotation)
        except ValueError:
            raise IIIFRequestError(
                code=400, parameter="rotation",
                text="Bad rotation value, must be a number, got '%s'." %
                (self.rotation))
        if (self.rotation_deg < 0.0 or self.rotation_deg > 360.0):
            raise IIIFRequestError(
                code=400, parameter="rotation",
                text="Illegal rotation value, must be 0 <= rotation "
                     "<= 360, got %f." % (self.rotation_deg))
        elif (self.rotation_deg == 360.0):
            # The spec admits 360 as valid, but change to 0
            self.rotation_deg = 0.0

    def parse_quality(self):
        """Check quality paramater.

        Sets self.quality_val based on simple substitution of
        'native' for default. Checks for the three valid values
        else throws an IIIFRequestError.
        """
        if (self.quality is None):
            self.quality_val = self.default_quality
        elif (self.quality not in self.allowed_qualities):
            raise IIIFRequestError(
                code=400, parameter="quality",
                text="The quality parameter must be '%s', got '%s'." %
                ("', '".join(self.allowed_qualities), self.quality))
        else:
            self.quality_val = self.quality

    def parse_format(self):
        """Check format parameter.

        All formats values listed in the specification are lowercase
        alphanumeric value commonly used as file extensions. To leave
        opportunity for extension here just do a limited sanity check
        on characters and length.
        """
        if (self.format is not None and not re.match(r'''\w{1,20}$''', self.format)):
            raise IIIFRequestError(
                parameter='format',
                text='Bad format parameter')

    def __str__(self):
        """Return string of this object in human readable form.

        Distinguishes parametrerized and info requests to
        show only appropriate parameters in each case.
        """
        s = "baseurl=" + str(self.baseurl) + " "
        s += "identifier=" + str(self.identifier) + " "
        if (self.info):
            s += "INFO request "
            s += "format=" + str(self.format) + " "
        else:
            s += "region=" + str(self.region) + " "
            s += "size=" + str(self.size) + " "
            s += "rotation=" + str(self.rotation) + " "
            s += "quality=" + str(self.quality) + " "
            s += "format=" + str(self.format)
        return(s)



"""Base/null IIIF image manipulator.

Provides a number of utility methods to extract information necessary
for doing the transformations once one has knowledge of the source
image size.
"""

import logging
import os
import os.path
import re
import shutil
import subprocess

from .error import IIIFError, IIIFZeroSizeError
from .request import IIIFRequest


class IIIFManipulator(object):
    """Manipulate an image according to IIIF rules.

    All exceptions are raise as IIIFError objects which directly
    determine the HTTP response.
    """

    def __init__(self, api_version='2.1'):
        """Initialize Manipulator object.

        Accepts api_version as a parameter to tailor handling of
        requests according to the API being supported.

        Sets compliance_level to None because the null manipulator
        doesn't comply with any level. Sub-classes are expected to
        set this to a level number (0,1,2) appropriate to the
        facilities supported at the given API version. Use
        compliance_uri to get a URI.
        """
        self.api_version = api_version
        self.compliance_level = None
        self.max_area = None
        self.max_width = None
        self.max_height = None
        self.srcfile = None
        self.request = None
        self.outfile = None
        self.logger = logging.getLogger(__name__)

    @property
    def compliance_uri(self):
        """Compliance URI based on api_version.

        Value is based on api_version and complicance_level, will be
        None if either are unset/unrecognized. The assumption here is
        that the api_version and level are orthogonal, override this
        method if that isn't true.
        """
        if (self.api_version == '1.0'):
            uri_pattern = r'http://library.stanford.edu/iiif/image-api/compliance.html#level%d'
        elif (self.api_version == '1.1'):
            uri_pattern = r'http://library.stanford.edu/iiif/image-api/1.1/compliance.html#level%d'
        elif (self.api_version == '2.0' or
              self.api_version == '2.1'):
            uri_pattern = r'http://iiif.io/api/image/2/level%d.json'
        else:
            return
        if (self.compliance_level is None):
            return
        return(uri_pattern % self.compliance_level)

    def derive(self, srcfile=None, request=None, outfile=None):
        """Do sequence of manipulations for IIIF to derive output image.

        Named argments:
        srcfile -- source image file
        request -- IIIFRequest object with parsed parameters
        outfile -- output image file. If set the the output file will be
                   written to that file, otherwise a new temporary file
                   will be created and outfile set to its location.

        See order in spec: http://www-sul.stanford.edu/iiif/image-api/#order

          Region THEN Size THEN Rotation THEN Quality THEN Format

        Typical use:

            r = IIIFRequest(region=...)
            m = IIIFManipulator()
            try:
                m.derive(srcfile='a.jpg',request=r)
                # .. serve m.outfile
            except IIIFError as e:
                # ..
            finally:
                m.cleanup() #removes temp m.outfile
        """
        # set if specified
        if (srcfile is not None):
            self.srcfile = srcfile
        if (request is not None):
            self.request = request
        if (outfile is not None):
            self.outfile = outfile
        if (self.outfile is not None):
            # create path to output dir if necessary
            dir = os.path.dirname(self.outfile)
            if (not os.path.exists(dir)):
                os.makedirs(dir)
        #
        self.do_first()
        (x, y, w, h) = self.region_to_apply()
        self.do_region(x, y, w, h)
        (w, h) = self.size_to_apply()
        self.do_size(w, h)
        (mirror, rot) = self.rotation_to_apply(no_mirror=True)
        self.do_rotation(mirror, rot)
        (quality) = self.quality_to_apply()
        self.do_quality(quality)
        self.do_format(self.request.format)
        self.do_last()
        return(self.outfile, self.mime_type)

    def do_first(self):
        """Simplest possible manipulator that can only handle no modification.

        Set width and height to -1 (unknown).
        """
        self.width = -1  # don't know width of height
        self.height = -1

    def do_region(self, x, y, w, h):
        """Null implementation of region selection."""
        if (x is not None):
            raise IIIFError(code=501, parameter="region",
                            text="Null manipulator supports only region=/full/.")

    def do_size(self, w, h):
        """Null implementation of size scaling."""
        if (w is not None):
            raise IIIFError(code=501, parameter="size",
                            text="Null manipulator supports only size=pct:100 and size=full.")

    def do_rotation(self, mirror, rot):
        """Null implementation of rotate and/or mirror."""
        if (mirror):
            raise IIIFError(code=501, parameter="rotation",
                            text="Null manipulator does not support mirroring.")
        if (rot != 0.0):
            raise IIIFError(code=501, parameter="rotation",
                            text="Null manipulator supports only rotation=(0|360).")

    def do_quality(self, quality):
        """Null implementation of quality."""
        if (self.api_version >= '2.0'):
            if (quality != "default"):
                raise IIIFError(code=501, parameter="default",
                                text="Null manipulator supports only quality=default.")
        else:  # versions 1.0 and 1.1
            if (quality != "native"):
                raise IIIFError(code=501, parameter="native",
                                text="Null manipulator supports only quality=native.")

    def do_format(self, format):
        """Null implementation of format selection.

        This is the last step, this null implementation does not accept any
        specification of a format because we don't even know what the input
        format is.
        """
        if (format is not None):
            raise IIIFError(code=415, parameter="format",
                            text="Null manipulator does not support specification of output format.")
        #
        if (self.outfile is None):
            self.outfile = self.srcfile
        else:
            try:
                shutil.copyfile(self.srcfile, self.outfile)
            except IOError as e:
                raise IIIFError(code=500,
                                text="Failed to copy file (%s)." % (str(e)))
        self.mime_type = None

    def do_last(self):
        """Null implementation of hook in pipeline at end of processing.

        Does nothing.
        """
        return

    # Utility methods

    def region_to_apply(self):
        """Return the x,y,w,h parameters to extract given image width and height.

        Assume image width and height are available in self.width and
        self.height, and self.request is IIIFRequest object

        Expected use:
          (x,y,w,h) = self.region_to_apply()
          if (x is None):
              # full image
          else:
              # extract

        Returns (None,None,None,None) if no extraction is required.
        """
        if (self.request.region_full or
            (self.request.region_pct and
             self.request.region_xywh == (0, 0, 100, 100))):
            return(None, None, None, None)
        # Cannot do anything else unless we know size (in self.width and
        # self.height)
        if (self.width <= 0 or self.height <= 0):
            raise IIIFError(code=501, parameter='region',
                            text="Region parameters require knowledge of image size which is not implemented.")
        if (self.request.region_square):
            if (self.width <= self.height):
                y_offset = (self.height - self.width) / 2
                return(0, y_offset, self.width, self.width)
            else:  # self.width>self.height
                x_offset = (self.width - self.height) / 2
                return(x_offset, 0, self.height, self.height)
        # pct or explicit pixel sizes
        pct = self.request.region_pct
        (x, y, w, h) = self.request.region_xywh
        # Convert pct to pixels based on actual size
        if (pct):
            x = int((x / 100.0) * self.width + 0.5)
            y = int((y / 100.0) * self.height + 0.5)
            w = int((w / 100.0) * self.width + 0.5)
            h = int((h / 100.0) * self.height + 0.5)
        # Check if boundary extends beyond image and truncate
        if ((x + w) > self.width):
            w = self.width - x
        if ((y + h) > self.height):
            h = self.height - y
        # Final check to see if we have the whole image
        if (w == 0 or h == 0):
            raise IIIFZeroSizeError(code=400, parameter='region',
                                    text="Region parameters would result in zero size result image.")
        if (x == 0 and y == 0 and w == self.width and h == self.height):
            return(None, None, None, None)
        return(x, y, w, h)

    def size_to_apply(self):
        """Calculate size of image scaled using size parameters.

        Assumes current image width and height are available in self.width and
        self.height, and self.request is IIIFRequest object.

        Formats are: w, ,h w,h pct:p !w,h full max

        Returns (None,None) if no scaling is required.

        If max is requested and neither max_area or max_width are
        specified then this is the same as full. Otherwise the limits
        are used to determine the size.
        """
        if (self.request.size_full or self.request.size_pct == 100.0):
            # full size
            return(None, None)
        # Not trivially full size, look at possibilities in turn
        w = self.width
        h = self.height
        if (self.request.size_max):
            # use size limits if present, else full
            if (self.max_area and self.max_area < (w * h)):
                scale = (float(self.max_area) / float(w * h)) ** 0.5
                w = int(w * scale + 0.5)
                h = int(h * scale + 0.5)
            if (self.max_width):
                max_height = self.max_height if self.max_height is not None else self.max_width
                if (self.max_width < w):
                    # calculate wrt original width, height rather than
                    # w, h to avoid compounding rounding issues
                    scale = float(self.max_width) / float(self.width)
                    w = int(self.width * scale + 0.5)
                    h = int(self.height * scale + 0.5)
                if (max_height < h):
                    scale = float(max_height) / float(self.height)
                    w = int(self.width * scale + 0.5)
                    h = int(self.height * scale + 0.5)
        elif (self.request.size_pct is not None):
            w = int(self.width * self.request.size_pct / 100.0 + 0.5)
            h = int(self.height * self.request.size_pct / 100.0 + 0.5)
        elif (self.request.size_bang):
            # Have "!w,h" form
            (mw, mh) = self.request.size_wh
            # Pick smaller fraction and then work from that...
            frac = min((float(mw) / float(self.width)),
                       (float(mh) / float(self.height)))
            w = int(self.width * frac + 0.5)
            h = int(self.height * frac + 0.5)
        else:
            # Must now be "w,h", "w," or ",h". If both are specified then this will the size,
            # otherwise find other to keep aspect ratio
            (w, h) = self.request.size_wh
            if (w is None):
                w = int(self.width * h / self.height + 0.5)
            elif (h is None):
                h = int(self.height * w / self.width + 0.5)
        # Now have w,h, sanity check and return
        if (w == 0 or h == 0):
            raise IIIFZeroSizeError(
                code=400, parameter='size',
                text="Size parameter would result in zero size result image (%d,%d)." % (w, h))
        # Below would be test for scaling up image size, this is allowed by spec
        # if ( w>self.width or h>self.height ):
        #      raise IIIFError(code=400,parameter='size',
        # text="Size requests scaling up image to larger than orginal.")
        if (w == self.width and h == self.height):
            return(None, None)
        return(w, h)

    def rotation_to_apply(self, only90s=False, no_mirror=False):
        """Check and interpret rotation.

        Returns a truth value as to whether to mirror, and a floating point
        number 0 <= angle < 360 (degrees).
        """
        rotation = self.request.rotation_deg
        if (no_mirror and self.request.rotation_mirror):
            raise IIIFError(code=501, parameter="rotation",
                            text="This implementation does not support mirroring.")
        if (only90s and (rotation != 0.0 and rotation != 90.0 and
                         rotation != 180.0 and rotation != 270.0)):
            raise IIIFError(code=501, parameter="rotation",
                            text="This implementation supports only 0,90,180,270 degree rotations.")
        return(self.request.rotation_mirror, rotation)

    def quality_to_apply(self):
        """Value of quality parameter to use in processing request.

        Simple substitution of 'native' or 'default' if no quality
        parameter is specified.
        """
        if (self.request.quality is None):
            if (self.api_version <= '1.1'):
                return('native')
            else:
                return('default')
        return(self.request.quality)

    def cleanup(self):
        """Null implementation of clean up after derive call and use of output.

        Call after any output file from the derivation process has been
        read. Intended to handle any clean up of temporary files or such.
        This method empty in base class.
        """
        pass

    def scale_factors(self, tile_width, tile_height=None):
        """Return a set of scale factors for given tile and window size.

        Gives a set of scale factors, starting at 1, and in multiples
        of 2. Largest scale_factor is so that one tile will cover the
        entire image (self.width,self.height).

        If tile_height is not specified then tiles are assumed to be
        squares of tile_width pixels.
        """
        if (not tile_height):
            tile_height = tile_width
        sf = 1
        scale_factors = [sf]
        for j in range(30):  # limit of 2^30, should be enough!
            sf = 2 * sf
            if (tile_width * sf > self.width and
                    tile_height * sf > self.height):
                break
            scale_factors.append(sf)
        return scale_factors



"""Support for IIIF Authentication."""

import json
import logging
import random
import re


class IIIFAuth(object):
    """IIIF Authentication Class.

    Base authentication class which does not actually implement an
    authentication scheme but is a foundation for specific sub-classes
    for different schemes.
    """

    def __init__(self, cookie_prefix=None):
        """Initialize IIIFAuth object."""
        self.profile_base = 'http://iiif.io/api/auth/0/'
        self.name = "image server"
        self.login_uri = None
        self.logout_uri = None
        self.client_id_uri = None
        self.access_token_uri = None
        self.logger = logging.getLogger(__name__)
        # Need to have different cookie names for each auth domain
        # running on the same server
        self.set_cookie_prefix(cookie_prefix)
        self.auth_cookie_name = self.cookie_prefix + 'loggedin'
        self.token_cookie_name = self.cookie_prefix + 'token'

    def set_cookie_prefix(self, cookie_prefix=None):
        """Set a random cookie prefix unless one is specified.

        In order to run multiple demonstration auth services on the
        same server we need to have different cookie names for each
        auth domain. Unless cookie_prefix is set, generate a random
        one.
        """
        if (cookie_prefix is None):
            self.cookie_prefix = "%06d_" % int(random.random() * 1000000)
        else:
            self.cookie_prefix = cookie_prefix

    def add_services(self, info):
        """Add auth service descriptions to an IIIFInfo object.

        Login service description is the wrapper for all other
        auth service descriptions so we have nothing unless
        self.login_uri is specified. I we do then add any other
        auth services at children.
        """
        if (self.login_uri):
            svc = self.login_service_description()
            svcs = []
            if (self.logout_uri):
                svcs.append(self.logout_service_description())
            if (self.client_id_uri):
                svcs.append(self.client_id_service_description())
            if (self.access_token_uri):
                svcs.append(self.access_token_service_description())
            # Add one as direct child of service property, else array for >1
            if (len(svcs) == 1):
                svc['service'] = svcs[0]
            elif (len(svcs) > 0):
                svc['service'] = svcs
            info.add_service(svc)

    def login_service_description(self):
        """Login service description."""
        return({"@id": self.login_uri,
                "profile": self.profile_base + 'login',
                "label": 'Login to ' + self.name})

    def logout_service_description(self):
        """Logout service description."""
        return({"@id": self.logout_uri,
                "profile": self.profile_base + 'logout',
                "label": 'Logout from ' + self.name})

    def client_id_service_description(self):
        """Client Id service description."""
        return({"@id": self.client_id_uri,
                "profile": self.profile_base + 'clientId'})

    def access_token_service_description(self):
        """Access Token service description."""
        return({"@id": self.access_token_uri,
                "profile": self.profile_base + 'token'})

    # Override with method to implement
    access_token_handler = None

    # Override with method to implement
    client_id_handler = None

    # Override with method to implement
    home_handler = None

    def info_authn(self):
        """Check to see if user is authenticated for info.json.

        Null implementation that always returns False, must override
        to implement authorization.
        """
        return False

    def info_authz(self):
        """Check to see if user is authenticated and authorized for info.json.

        Null implementation that says that any authenticated user is
        authorized.
        """
        return self.info_authn()

    def image_authn(self):
        """Check to see if user is authenticated for image requests.

        Null implementation that always returns False, must override
        to implement authorization.
        """
        return False

    def image_authz(self):
        """Check to see if user is authenticated and authorized for image requests.

        Null implementation that says that any authenticated user is
        authorized.
        """
        return self.image_authn()

    def scheme_host_port_prefix(self, scheme='http', host='host', port=None, prefix=None):
        """Return URI composed of scheme, server, port, and prefix."""
        uri = scheme + '://' + host
        if (port and not ((scheme == 'http' and port == 80) or
                          (scheme == 'https' and port == 443))):
            uri += ':' + str(port)
        if (prefix):
            uri += '/' + prefix
        return uri



"""IIIF Authentication using HTTP Basic auth.

FIXME - this code assumes Flask webapp framework, should be abstracted
"""

import json
import re
import os.path
from flask import request, make_response, redirect

from .auth import IIIFAuth


class IIIFAuthBasic(IIIFAuth):
    """IIIF Authentication Class using HTTP Basic Auth."""

    def __init__(self, cookie_prefix=None):
        """Initialize IIIFAuthBasic object."""
        super(IIIFAuthBasic, self).__init__(cookie_prefix=cookie_prefix)

    def logout_service_description(self):
        """Logout service description with has modified URI to supply 'bad' credentials.

        With HTTP Basic auth there are no clean ways to logout (except browser specific
        extensions). The standard recommendation seems to be to supply bad credentials.
        which we do in the URI (bad@X).
        """
        uri = re.sub(r'''^https?://''', lambda m: m.group(0) +
                     'bad:X@', self.logout_uri)
        return({"@id": uri,
                "profile": self.profile_base + 'logout',
                "label": 'Logout from ' + self.name
                })

    def info_authn(self):
        """Check to see if user if authenticated for info.json."""
        self.logger.info("info_authz: Authorization header = " +
                         request.headers.get('Authorization', '[none]'))
        # FIXME - check valaue
        return (request.headers.get('Authorization', '') != '')

    def image_authn(self):
        """Check to see if user if authenticated for image requests."""
        self.logger.info("image_authn: loggedin cookie = " +
                         request.cookies.get(self.token_cookie_name, default='[none]'))
        # FIXME - check value
        return request.cookies.get(self.token_cookie_name, default='')

    def login_handler(self, config=None, prefix=None, **args):
        """HTTP Basic login handler.

        Respond with 401 and WWW-Authenticate header if there are no credentials
        or bad credentials. If there are credentials then simply check for username
        equal to password for validity.
        """
        headers = {}
        headers['Access-control-allow-origin'] = '*'
        headers['Content-type'] = 'text/html'
        auth = request.authorization
        if (auth and auth.username == auth.password):
            html = "<html><script>window.close();</script></html>"
            response = make_response(html, 200, headers)
            response.set_cookie(self.auth_cookie_name,
                                "valid-http-basic-login", expires=3600)
            return response
        else:
            headers[
                'WWW-Authenticate'] = 'Basic realm="HTTP-Basic-Auth at %s (u=p to login)"' % (self.name)
            return make_response("", 401, headers)

    def logout_handler(self, **args):
        """Handler for logout button.

        Delete cookies and return HTML that immediately closes window. Logout
        is tricky but attempt to do this via self.logout_uri which as has
        bad username:password.
        """
        response = make_response(
            "<html><script>window.close();</script></html>", 200, {'Content-Type': "text/html"})
        response.set_cookie(self.auth_cookie_name, expires=0)
        response.set_cookie(self.token_cookie_name, expires=0)
        response.headers['Access-Control-Allow-Origin'] = '*'
        return response

    def access_token_handler(self, **args):
        """Get access token based on cookie sent with this request.

        The client requests a token to send in re-request for info.json.
        Support JSONP request to get the token to send to info.json in
        Auth'z header.
        """
        callback_function = request.args.get('callback', default='')
        authdone = request.headers.get('Authorization', '')
        if (authdone):
            token = 'secret_token_here'  # FIXME - need something real
            data = {"access_token": token,
                    "token_type": "Bearer", "expires_in": 3600}
        else:
            data = {"error": "client_unauthorized",
                    "error_description": "No login details received"}
        data_str = json.dumps(data)

        ct = "application/json"
        if (callback_function):
            data_str = "%s(%s);" % (callback_function, data_str)
            ct = "application/javascript"
        # Build response
        response = make_response(data_str, 200, {'Content-Type': ct})
        if (authdone):
            # Set the cookie for the image content -- FIXME - need something
            # real
            response.set_cookie(self.token_cookie_name, token)
        response.headers['Access-control-allow-origin'] = '*'
        return response



"""Setup for IIIF Image Library implementation."""
from setuptools import setup, Command
import os
# setuptools used instead of distutils.core so that
# dependencies can be handled automatically

# Extract version number from iiif/_version.py. Here we are very strict
# about the format of the version string as an extra sanity check.
# (thanks for comments in
# http://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
# )
import re
VERSIONFILE = "iiif/_version.py"
verfilestr = open(VERSIONFILE, "rt").read()
match = re.search(r"^__version__ = '(\d\.\d.\d+(\.\d+)?)'",
                  verfilestr, re.MULTILINE)
if match:
    version = match.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE))


class Coverage(Command):
    """Class to allow coverage run from setup."""

    description = "run coverage"
    user_options = []

    def initialize_options(self):
        """Empty initialize_options."""
        pass

    def finalize_options(self):
        """Empty finalize_options."""
        pass

    def run(self):
        """Run coverage program."""
        os.system(
            "coverage run --source=iiif --omit=iiif/manipulator_netpbm.py setup.py test")
        os.system("coverage report")
        os.system("coverage html")
        print("See htmlcov/index.html for details.")

setup(
    name='iiif',
    version=version,
    author='Simeon Warner',
    author_email='simeon.warner@cornell.edu',
    packages=['iiif'],
    package_data={'iiif': ['templates/*',
                           'third_party/jquery-1.11.1.min.js',
                           'third_party/openseadragon121/*.js',
                           'third_party/openseadragon121/images/*',
                           'third_party/openseadragon100/*.js',
                           'third_party/openseadragon100/images/*',
                           'third_party/openseadragon200/*.js',
                           'third_party/openseadragon200/images/*']},
    data_files=['demo-static/README.md'],
    scripts=['iiif_static.py', 'iiif_testserver.py'],
    classifiers=["Development Status :: 4 - Beta",
                 "Intended Audience :: Developers",
                 "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                 "Operating System :: OS Independent",  # is this true? know Linux & OS X ok
                 "Programming Language :: Python",
                 "Programming Language :: Python :: 2.6",
                 "Programming Language :: Python :: 2.7",
                 "Programming Language :: Python :: 3.3",
                 "Programming Language :: Python :: 3.4",
                 "Topic :: Internet :: WWW/HTTP",
                 "Topic :: Multimedia :: Graphics :: Graphics Conversion",
                 "Topic :: Software Development :: Libraries :: Python Modules",
                 "Environment :: Web Environment"],
    url='https://github.com/zimeon/iiif',
    license='LICENSE.txt',
    description='IIIF Image API reference implementation',
    long_description=open('README').read(),
    install_requires=[
        "Pillow",
        "python-magic",
        "Flask",
        "mock"
    ],
    test_suite="tests",
    tests_require=[
        "testfixtures",
    ],
    cmdclass={
        'coverage': Coverage,
    },
)



"""Static file generation for IIIF Image API.

Use IIIF Image API manipulations to generate a set of tiles for
a level0 implementation of the IIIF Image API using static files.
"""

import math
import logging
import os
import os.path
import shutil
from string import Template

from .manipulator_pil import IIIFManipulatorPIL
from .manipulator_gen import IIIFManipulatorGen
from .info import IIIFInfo
from .request import IIIFRequest
from .error import IIIFZeroSizeError


def static_partial_tile_sizes(width, height, tilesize, scale_factors):
    """Generator for partial tile sizes for zoomed in views.

    Positional arguments:
    width -- width of full size image
    height -- height of full size image
    tilesize -- width and height of tiles
    scale_factors -- iterable of scale factors, typically [1,2,4..]

    Yields ([rx,ry,rw,rh],[sw,sh]), the region and size for each tile
    """
    for sf in scale_factors:
        if (sf * tilesize >= width and sf * tilesize >= height):
            continue  # avoid any full-region tiles
        rts = tilesize * sf  # tile size in original region
        xt = (width - 1) // rts + 1
        yt = (height - 1) // rts + 1
        for nx in range(xt):
            rx = nx * rts
            rxe = rx + rts
            if (rxe > width):
                rxe = width
            rw = rxe - rx
            # same as sw = int(math.ceil(rw/float(sf)))
            sw = (rw + sf - 1) // sf
            for ny in range(yt):
                ry = ny * rts
                rye = ry + rts
                if (rye > height):
                    rye = height
                rh = rye - ry
                # same as sh = int(math.ceil(rh/float(sf)))
                sh = (rh + sf - 1) // sf
                yield([rx, ry, rw, rh], [sw, sh])


def static_full_sizes(width, height, tilesize):
    """Generator for scaled-down full image sizes.

    Positional arguments:
    width -- width of full size image
    height -- height of full size image
    tilesize -- width and height of tiles

    Yields [sw,sh], the size for each full-region tile that is less than
    the tilesize. This includes tiles up to the full image size if that
    is smaller than the tilesize.
    """
    # FIXME - Not sure what correct algorithm is for this, from
    # observation of Openseadragon it seems that one keeps halving
    # the pixel size of the full image until until both width and
    # height are less than the tile size. After that all subsequent
    # halving of the image size are used, all the way down to 1,1.
    # It seems that without these reduced size full-region images,
    # OpenSeadragon will not display any unzoomed image in small windows.
    #
    # I do not understand the algorithm that OpenSeadragon uses (or
    # know where it is in the code) to decide how small a version of
    # the complete image to request. It seems that there is a bug in
    # OpenSeadragon here because in some cases it requests images
    # of size 1,1 multiple times, which is anyway a useless image.
    for level in range(0, 20):
        factor = 2.0**level
        sw = int(width / factor + 0.5)
        sh = int(height / factor + 0.5)
        if (sw < tilesize and sh < tilesize):
            if (sw < 1 or sh < 1):
                break
            yield([sw, sh])


class IIIFStaticError(Exception):
    """Error class for errors to be resported to user."""

    pass


class IIIFStatic(object):
    """Provide static generation of IIIF images.

    Simplest, using source image as model for directory which
    will be in same directory without extension:

        IIIFStatic("image1.jpg").generate()

    More complex with different output directory but using name
    derived from image:

        sg = IIIFStatic(dst="outdir")
        sg.generate("image2.jpg")
        sg.generate("image3.jpg")

    The class is quite noisy at level logging.INFO, set the logging
    level to logging.WARNING to get log output only when there are
    warnings or errors.
    """

    def __init__(self, src=None, dst=None, tilesize=None,
                 api_version='2.0', dryrun=None, prefix='',
                 osd_version=None, generator=False,
                 max_image_pixels=0):
        """Initialization for IIIFStatic instances.

        All keyword arguments are optional:
        src -- source image file
        dst -- destination directory
        tilesize -- size of square tiles to generate (default 512)
        api_version -- IIIF Image API version to support (default 2.0)
        dryrun -- True to not write any output (default None)
        prefix -- identifier prefix
        osd_version -- use a specific version of OpenSeadragon
        """
        self.src = src
        self.dst = dst
        self.tilesize = tilesize if tilesize is not None else 512
        self.api_version = api_version
        if (self.api_version == '1'):
            self.api_version = '1.1'
        elif (self.api_version == '2'):
            self.api_version = '2.0'
        self.dryrun = (dryrun is not None)
        self.logger = logging.getLogger(__name__)
        self.prefix = prefix
        self.osd_version = osd_version if osd_version else '2.0.0'
        if (generator):
            self.manipulator_klass = IIIFManipulatorGen
        else:
            self.manipulator_klass = IIIFManipulatorPIL
        self.max_image_pixels = max_image_pixels
        # config for locations of OpenSeadragon
        # - dir is relative to base, will be copied to dir under html_dir
        # - js and images are relative to dir
        self.module_dir = os.path.dirname(__file__)
        self.osd_config = {
            '1.0.0': {
                'base': os.path.join(self.module_dir, 'third_party'),
                'dir': 'openseadragon100',
                'js': 'openseadragon.min.js',
                'images': 'images',
                'use_canonical': False,
                'notes': 'Uses /w,h/ for size, algorithm not quite right, API v1.1'
            },
            '1.2.1': {
                'base': os.path.join(self.module_dir, 'third_party'),
                'dir': 'openseadragon121',
                'js': 'openseadragon.min.js',
                'images': 'images',
                'use_canonical': True,
                'notes': "Uses /w,/ canonical syntax, works with API v1.1,2.0"
            },
            '2.0.0': {
                'base': os.path.join(self.module_dir, 'third_party'),
                'dir': 'openseadragon200',
                'js': 'openseadragon.min.js',
                'images': 'images',
                'use_canonical': True,
                'notes': "Uses /w,/ canonical syntax, works with API v1.1,2.0"
            }
        }
        # used internally:
        self.identifier = None
        self.copied_osd = False
        self.template_dir = os.path.join(self.module_dir, 'templates')

    def get_osd_config(self, osd_version):
        """Select appropriate portion of config.

        If the version requested is not supported the raise an exception with
        a helpful error message listing the versions supported.
        """
        if (osd_version in self.osd_config):
            return(self.osd_config[osd_version])
        else:
            raise IIIFStaticError("OpenSeadragon version %s not supported, available versions are %s" %
                                  (osd_version, ', '.join(sorted(self.osd_config.keys()))))

    def generate(self, src=None, identifier=None):
        """Generate static files for one source image."""
        self.src = src
        self.identifier = identifier
        # Get image details and calculate tiles
        im = self.manipulator_klass()
        im.srcfile = self.src
        im.set_max_image_pixels(self.max_image_pixels)
        im.do_first()
        width = im.width
        height = im.height
        scale_factors = im.scale_factors(self.tilesize)
        # Setup destination and IIIF identifier
        self.setup_destination()
        # Write out images
        for (region, size) in static_partial_tile_sizes(width, height, self.tilesize, scale_factors):
            self.generate_tile(region, size)
        sizes = []
        for (size) in static_full_sizes(width, height, self.tilesize):
            # See https://github.com/zimeon/iiif/issues/9
            sizes.append({'width': size[0], 'height': size[1]})
            self.generate_tile('full', size)
        # Write info.json
        qualities = ['default'] if (self.api_version > '1.1') else ['native']
        info = IIIFInfo(level=0, server_and_prefix=self.prefix, identifier=self.identifier,
                        width=width, height=height, scale_factors=scale_factors,
                        tile_width=self.tilesize, tile_height=self.tilesize,
                        formats=['jpg'], qualities=qualities, sizes=sizes,
                        api_version=self.api_version)
        json_file = os.path.join(self.dst, self.identifier, 'info.json')
        if (self.dryrun):
            self.logger.warning(
                "dryrun mode, would write the following files:")
            self.logger.warning("%s / %s/%s" %
                                (self.dst, self.identifier, 'info.json'))
        else:
            with open(json_file, 'w') as f:
                f.write(info.as_json())
                f.close()
            self.logger.info("%s / %s/%s" %
                             (self.dst, self.identifier, 'info.json'))
            self.logger.debug("Written %s" % (json_file))

    def generate_tile(self, region, size):
        """Generate one tile for this given region,size of this region.

        Logically we might use `w,h` instead of the Image API v2.0 canonical
        form `w,` if the api_version is 1.x. However, OSD 1.2.1 and 2.x assume
        the new canonical form even in the case where the API version is declared
        earlier. Thus, determine whether to use the canonical or `w,h` form based
        solely on the setting of osd_version.
        """
        r = IIIFRequest(identifier=self.identifier,
                        api_version=self.api_version)
        osd_config = self.get_osd_config(self.osd_version)
        use_canonical = osd_config['use_canonical']
        if (region == 'full'):
            r.region_full = True
        else:
            r.region_xywh = region  # [rx,ry,rw,rh]
        if (use_canonical):
            r.size_wh = [size[0], None]  # [sw,sh] -> [sw,]
        else:
            r.size_wh = size  # old form, full `w,h`
        r.format = 'jpg'
        path = r.url()
        # Generate...
        if (self.dryrun):
            self.logger.info("%s / %s" % (self.dst, path))
        else:
            m = self.manipulator_klass(api_version=self.api_version)
            try:
                m.derive(srcfile=self.src, request=r,
                         outfile=os.path.join(self.dst, path))
                self.logger.info("%s / %s" % (self.dst, path))
            except IIIFZeroSizeError:
                self.logger.info("%s / %s - zero size, skipped" %
                                 (self.dst, path))
                return()  # done if zero size
        if (region == 'full' and use_canonical):
            # In v2.0 of the spec, the canonical URI form `w,` for scaled
            # images of the full region was introduced. This is somewhat at
            # odds with the requirement for `w,h` specified in `sizes` to
            # be available, and has problems of precision with tall narrow
            # images. Hopefully will be fixed in 3.0 but for now symlink
            # the `w,h` form to the `w,` dirs so that might use the specified
            # `w,h` also work. See
            # <https://github.com/IIIF/iiif.io/issues/544>
            #
            # FIXME - This is ugly because we duplicate code in
            # iiif.request.url to construct the partial URL
            region_dir = os.path.join(r.quote(r.identifier), "full")
            wh_dir = "%d,%d" % (size[0], size[1])
            wh_path = os.path.join(region_dir, wh_dir)
            wc_dir = "%d," % (size[0])
            wc_path = os.path.join(region_dir, wc_dir)
            if (not self.dryrun):
                ln = os.path.join(self.dst, wh_path)
                if (os.path.exists(ln)):
                    os.remove(ln)
                os.symlink(wc_dir, ln)
            self.logger.info("%s / %s -> %s" % (self.dst, wh_path, wc_path))

    def setup_destination(self):
        """Setup output directory based on self.dst and self.identifier.

        Returns the output directory name on success, raises and exception on
        failure.
        """
        # Do we have a separate identifier?
        if (not self.identifier):
            # No separate identifier specified, split off the last path segment
            # of the source name, strip the extension to get the identifier
            self.identifier = os.path.splitext(os.path.split(self.src)[1])[0]
        # Done if dryrun, else setup self.dst first
        if (self.dryrun):
            return
        if (not self.dst):
            raise IIIFStaticError("No destination directory specified!")
        dst = self.dst
        if (os.path.isdir(dst)):
            # Exists, OK
            pass
        elif (os.path.isfile(dst)):
            raise IIIFStaticError(
                "Can't write to directory %s: a file of that name exists" % dst)
        else:
            os.makedirs(dst)
        # Second, create identifier based subdir if necessary
        outd = os.path.join(dst, self.identifier)
        if (os.path.isdir(outd)):
            # Nothing for now, perhaps should delete?
            self.logger.warning(
                "Output directory %s already exists, adding/updating files" % outd)
            pass
        elif (os.path.isfile(outd)):
            raise IIIFStaticError(
                "Can't write to directory %s: a file of that name exists" % outd)
        else:
            os.makedirs(outd)
        self.logger.debug("Output directory %s" % outd)

    def write_html(self, html_dir='/tmp', include_osd=False,
                   osd_width=500, osd_height=500):
        """Write HTML test page using OpenSeadragon for the tiles generated.

        Assumes that the generate(..) method has already been called to set up
        identifier etc. Parameters:
          html_dir - output directory for HTML files, will be created if it
                     does not already exist
          include_osd - true to include OpenSeadragon code
          osd_width - width of OpenSeadragon pane in pixels
          osd_height - height of OpenSeadragon pane in pixels
        """
        osd_config = self.get_osd_config(self.osd_version)
        osd_base = osd_config['base']
        osd_dir = osd_config['dir']  # relative to base
        osd_js = os.path.join(osd_dir, osd_config['js'])
        osd_images = os.path.join(osd_dir, osd_config['images'])
        if (os.path.isdir(html_dir)):
            # Exists, fine
            pass
        elif (os.path.isfile(html_dir)):
            raise IIIFStaticError(
                "Can't write to directory %s: a file of that name exists" % html_dir)
        else:
            os.makedirs(html_dir)
        self.logger.info("Writing HTML to %s" % (html_dir))
        with open(os.path.join(self.template_dir, 'static_osd.html'), 'r') as f:
            template = f.read()
        outfile = self.identifier + '.html'
        outpath = os.path.join(html_dir, outfile)
        with open(outpath, 'w') as f:
            info_json_uri = '/'.join([self.identifier, 'info.json'])
            if (self.prefix):
                info_json_uri = '/'.join([self.prefix, info_json_uri])
            d = dict(identifier=self.identifier,
                     api_version=self.api_version,
                     osd_version=self.osd_version,
                     osd_uri=osd_js,
                     osd_images_prefix=osd_images,
                     osd_height=osd_width,
                     osd_width=osd_height,
                     info_json_uri=info_json_uri)
            f.write(Template(template).safe_substitute(d))
            self.logger.info("%s / %s" % (html_dir, outfile))
        # Do we want to copy OSD in there too? If so, do it only if
        # we haven't already
        if (include_osd):
            if (self.copied_osd):
                self.logger.info("OpenSeadragon already copied")
            else:
                # Make directory, copy JavaScript and icons (from osd_images)
                osd_path = os.path.join(html_dir, osd_dir)
                if (not os.path.isdir(osd_path)):
                    os.makedirs(osd_path)
                shutil.copyfile(os.path.join(osd_base, osd_js),
                                os.path.join(html_dir, osd_js))
                self.logger.info("%s / %s" % (html_dir, osd_js))
                osd_images_path = os.path.join(html_dir, osd_images)
                if (os.path.isdir(osd_images_path)):
                    self.logger.warning(
                        "OpenSeadragon images directory (%s) already exists, skipping"
                        % osd_images_path)
                else:
                    shutil.copytree(os.path.join(osd_base, osd_images),
                                    osd_images_path)
                    self.logger.info("%s / %s/*" % (html_dir, osd_images))
                self.copied_osd = True  # don't try again for next img
